"""
Utility methods to print system and dependency information.

adapted from :func:`pandas.show_versions` (`locan/licences/PANDAS.rst`)
(BSD 3-Clause License, Copyright (c) 2008-2011, AQR Capital Management, LLC, Lambda Foundry, Inc. and PyData Development
Team, Copyright (c) 2011-2021, Open source contributors.)
and from :func:`scikit-learn.show_versions` (`locan/licences/SCIKIT-LEARN.rst`)
(BSD 3-Clause License, Copyright (c) 2007-2020 The scikit-learn developers.)
"""

import platform
import sys
import importlib
import struct
import os
import locale

from locan.dependencies import INSTALL_REQUIRES, EXTRAS_REQUIRE, IMPORT_NAMES
from locan import __version__ as locan_version


__all__ = ['system_info', 'dependency_info', 'show_versions']


def system_info(verbose=True):
    """
    Return system information.

    Parameters
    ----------
    verbose : bool
        If True information on node and executable path are added.

    Return
    ------
    dict
        System information
    """
    uname_result = platform.uname()
    language_code, encoding = locale.getlocale()

    sys_info = {
        "python-bits": struct.calcsize("P") * 8,
        "system": uname_result.system,
        "release": uname_result.release,
        "version": uname_result.version,
        "machine": uname_result.machine,
        "processor": uname_result.processor,
        "byteorder": sys.byteorder,
        "LC_ALL": os.environ.get("LC_ALL"),
        "LANG": os.environ.get("LANG"),
        "LOCALE": {"language-code": language_code, "encoding": encoding},
    }

    if verbose:
        sys_info.update({
            "node": uname_result.node,
            "executable": sys.executable,
        })

    return sys_info


def dependency_info(extra_dependencies=True, other_dependencies=None):
    """
    Overview of the installed version of main dependencies.

    Parameters
    ----------
    extra_dependencies : bool
        Include extra dependencies as specified in setup.py

    other_dependencies : list or None
        Include other module names.

    Returns
    -------
    dict
        Version information on relevant Python libraries
    """
    deps = INSTALL_REQUIRES.copy()

    if extra_dependencies:
        deps = deps.union(EXTRAS_REQUIRE)

    if other_dependencies:
        deps = deps.union(other_dependencies)

    deps_info = {}
    for modname in deps:
        try:
            deps_info[modname] = importlib.metadata.version(modname)
        except importlib.metadata.PackageNotFoundError:
            deps_info[modname] = None
    return deps_info


def show_versions(locan=True, python=True, system=True, dependencies=True,
                  verbose=True, extra_dependencies=True, other_dependencies=None):
    """
    Print useful debugging information on system and dependency versions.

    Parameters
    ----------
    locan : bool
        Show locan version
    python : bool
        Show python version
    system : bool
        Show system information
    verbose : bool
        If True information on node and executable path are added.
    dependencies : bool
        Show main dependencies
    extra_dependencies : bool
        Include extra dependencies as specified in setup.py if True.
    other_dependencies : list or None
        Include other module names.

    Returns
    -------
    None
    """

    locan_info = {'version': locan_version}
    python_info = {'version': platform.python_version()}
    sys_info = system_info(verbose)
    deps_info = dependency_info(extra_dependencies, other_dependencies)

    if locan:
        print('\nLocan:')
        for key, value in locan_info.items():
            print(f"{key:>10}: {value}")

    if python:
        print('\nPython:')
        for key, value in python_info.items():
            print(f"{key:>10}: {value}")

    if system:
        print('\nSystem:')
        for key, value in sys_info.items():
            print(f"{key:>10}: {value}")

    if dependencies:
        print('\nPython dependencies:')
        for key, value in deps_info.items():
            print(f"{key:>10}: {value}")



import pytest
import numpy as np
import pandas as pd
from locan import LocData
from locan.analysis import LocalizationUncertaintyFromIntensity


@pytest.fixture()
def locdata_simple():
    locdata_dict = {
        'position_x': [0, 0, 1, 4, 5],
        'position_y': [0, 1, 3, 4, 1],
        'intensity': [0, 1, 3, 4, 1],
        'psf_sigma_x': [100, 100, 100, 100, 100],
        }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


def test_uncertainty_empty(caplog):
    unc = LocalizationUncertaintyFromIntensity().compute(LocData())
    assert caplog.record_tuples == [('locan.analysis.uncertainty', 30, 'Locdata is empty.')]


def test_uncertainty(locdata_simple):
    unc = LocalizationUncertaintyFromIntensity().compute(locdata_simple)
    # print(unc.results)
    # print(unc.results['Uncertainty_x'][0])
    assert unc.results['uncertainty_x'][0] == np.inf
    assert unc.results['uncertainty_x'][1] == 100



"""

Regions as support for localization data.

This module provides classes to define geometric regions for localization data.
All region classes inherit from the abstract base class `Region`.

"""
# todo: fix docstrings

from abc import ABC, abstractmethod
import itertools as it

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.path as mpl_path
import matplotlib.patches as mpl_patches
from scipy.spatial.distance import pdist
from shapely.geometry import Point as shPoint
from shapely.geometry import MultiPoint as shMultiPoint
from shapely.geometry import Polygon as shPolygon
from shapely.geometry import MultiPolygon as shMultiPolygon
from shapely.prepared import prep
from shapely.affinity import scale, rotate, translate


__all__ = ['Region', 'EmptyRegion',
           'Region1D', 'Interval',
           'Region2D', 'Rectangle', 'Ellipse', 'Polygon', 'MultiPolygon',
           'Region3D', 'AxisOrientedCuboid', 'Cuboid',
           'RegionND', 'AxisOrientedHypercuboid'
           ]

# __all__ += ['Ellipsoid' 'Polyhedron']
# __all__ += ['Polytope']

__all__ += ['RoiRegion']  # legacy code that is only needed for legacy _roi.yml files.


class RoiRegion:
    """
    Deprecated Region object to specify regions of interest.

    A region that defines a region of interest with methods for getting a printable representation (that can also be
    saved in a yaml file), for returning a matplotlib patch that can be shown in a graph, for finding points within
    the region.

    Warnings
    ________
    This class is to be deprecated and should only be used to deal with legacy _roi.yaml files.
    Use Region classes instead.

    Parameters
    ----------
    region_type : str
        A string indicating the roi shape.
        In 1D it can be `interval`.
        In 2D it can be either `rectangle`, `ellipse`, or closed `polygon`.
        In 2D it can also be `shapelyPolygon` or `shapelyMultiPolygon`.
        In 3D it can be either `cuboid` or `ellipsoid` or `polyhedron` (not implemented yet).
    region_specs : tuple
        1D rois are defined by the following tuple:
        * interval: (start, stop)
        2D rois are defined by the following tuples:
        * rectangle: ((corner_x, corner_y), width, height, angle) with angle in degree
        * ellipse: ((center_x, center_y), width, height, angle) with angle in degree
        * polygon: ((point1_x, point1_y), (point2_x, point2_y), ..., (point1_x, point1_y))
        * shapelyPolygon: ((point_tuples), ((hole_tuples), ...))
        * shapelyMultiPolygon: (shapelyPolygon_specs_1, shapelyPolygon_specs_2, ...)
        3D rois are defined by the following tuples:
        * cuboid: ((corner_x, corner_y, corner_z), length, width, height, angle_1, angle_2, angle_3)
        * ellipsoid: ((center_x, center_y, center_z), length, width, height, angle_1, angle_2, angle_3)
        * polyhedron: (...)

    Attributes
    ----------
    region_type : str
        Type of region
    region_specs : tuple
        Specifications for region
    _region : RoiRegion
        RoiRegion instance for the specified region type.
    polygon : numpy.ndarray of tuples
        Array of points for a closed polygon approximating the region of interest in clockwise orientation. The first
        and last point must be identical.
    dimension : int
        Spatial dimension of region
    centroid : tuple of float
        Centroid coordinates
    max_distance : array-like of float
        Maximum distance between any two points in the region
    region_measure : float
        Hull measure, i.e. area or volume
    subregion_measure : float
        Measure of the sub-dimensional region, i.e. circumference or surface.
    """

    def __init__(self, region_type, region_specs):
        self.region_specs = region_specs
        self.region_type = region_type

        if region_type == 'interval':
            self._region = Interval(*region_specs)

        elif region_type == 'rectangle':
            self._region = Rectangle(*region_specs)

        elif region_type == 'ellipse':
            self._region = Ellipse(*region_specs)

        elif region_type == 'polygon':
            self._region = Polygon(region_specs)

        elif region_type == 'shapelyPolygon':
            self._region = Polygon(region_specs)

        elif region_type == 'shapelyMultiPolygon':
            self._region = MultiPolygon(region_specs)

        else:
            raise NotImplementedError(f'Region_type {region_type} has not been implemented yet.')

    def __getattr__(self, attr):
        """All non-adapted calls are passed to the _region object"""
        if attr.startswith('__') and attr.endswith('__'):  # this is needed to enable pickling
            raise AttributeError
        return getattr(self._region, attr)

    def __repr__(self):
        """
        Readable, printable and savable representation of RoiRegion.
        """
        return str(dict(region_type=self.region_type, region_specs=self.region_specs))

    @property
    def region(self):
        return self._region

    @classmethod
    def from_shapely(cls, region_type, shapely_obj):
        if region_type == 'shapelyPolygon':
            region_specs = Polygon.from_shapely(shapely_obj).region_specs

        elif region_type == 'shapelyMultiPolygon':
            region_specs = MultiPolygon.from_shapely(shapely_obj).region_specs

        else:
            raise NotImplementedError(f'Region_type {region_type} has not been implemented yet.')

        return cls(region_type=region_type, region_specs=region_specs)

    def contains(self, points):
        """
        Return list of indices for all points that are inside the region of interest.

        Parameters
        ----------
        points : array
            2D or 3D coordinates of oints that are tested for being inside the specified region.

        Returns
        -------
        numpy.ndarray of ints
            Array with indices for all points in original point array that are within the region.
        """
        return self._region.contains(points)

    def as_artist(self, **kwargs):
        """
        Matplotlib patch object for this region (e.g. `matplotlib.patches.Ellipse`).

        Parameters
        ----------
        kwargs : dict
            Other parameters passed to the `matplotlib.patches` object.

        Returns
        -------
        patch : matplotlib.patches
            Matplotlib patch for the specified region.
        """
        return self._region.as_artist(**kwargs)

    def to_shapely(self):
        """
        Convert region to a polygon and return as shapely object.

        Returns
        -------
        shapely.Polygon
        """
        return self._region.shapely_object


class Region(ABC):
    """
       Abstract Region class to define the interface for Region-derived classes that specify geometric objects
       to represent regions of interest.
    """
    def __repr__(self):
        return f'{self.__class__.__name__}(...)'

    @classmethod
    def from_intervals(cls, intervals):
        """
        Constructor for instantiating Region from list of (min, max) bounds.
        Takes array-like intervals instead of interval to be consistent with `Rectangle.from_intervals`.

        Parameters
        ----------
        intervals : array-like of shape (2,)
            The region bounds for each dimension

        Returns
        -------
            Interval
        """
        if np.shape(intervals) == (2,):
            return Interval.from_intervals(intervals)
        elif np.shape(intervals) == (2, 2):
            return Rectangle.from_intervals(intervals)
        elif np.shape(intervals) == (3, 2):
            return AxisOrientedCuboid.from_intervals(intervals)
        elif np.shape(intervals)[0] > 3 and np.shape(intervals)[1] == 2:
            return AxisOrientedHypercuboid.from_intervals(intervals)
        else:
            raise TypeError("intervals must be of shape (2,) or (n_features, 2).")

    @property
    @abstractmethod
    def dimension(self):
        """
        The region dimension.

        Returns
        -------
        int
        """
        pass

    @property
    @abstractmethod
    def bounds(self):
        """
        Region bounds min_x, min_y, ..., max_x, max_y, ... for each dimension.

        Returns
        -------
        tuple of shape (2 * dimension,)
        """
        pass

    @property
    @abstractmethod
    def extent(self):
        """
        The extent (max_x - min_x), (max_y - min_y), ... for each dimension.

        Returns
        -------
        tuple or numpy-array of shape (dimension,)
        """
        pass

    @property
    @abstractmethod
    def points(self):
        """
        Point coordinates.

        Returns
        -------
        tuple or numpy-array of shape (n_points, dimension)
        """
        pass

    @property
    @abstractmethod
    def centroid(self):
        """
        Point coordinates for region centroid.

        Returns
        -------
        tuple or numpy-array of shape (dimension,)
        """
        pass

    @property
    @abstractmethod
    def max_distance(self):
        """
        The maximum distance between any two points within the region.

        Returns
        -------
        float
        """
        pass

    @property
    @abstractmethod
    def region_measure(self):
        """
        Region measure, i.e. area (for 2d) or volume (for 3d).

        Returns
        -------
        float
        """
        pass

    @property
    @abstractmethod
    def subregion_measure(self):
        """
        Measure of the sub-dimensional region, i.e. circumference (for 2d) or surface (for 3d).

        Returns
        -------
        float
        """
        pass

    @property
    @abstractmethod
    def bounding_box(self):
        """
        A region describing the minimum axis-aligned bounding box that encloses the original region.

        Returns
        -------
        Region
        """
        pass

    @abstractmethod
    def intersection(self, other):
        """
        Returns a region representing the intersection of this region with
        ``other``.

        Parameters
        ----------
        other : Region
            Other region

        Returns
        -------
        Region
        """
        raise NotImplementedError

    @abstractmethod
    def symmetric_difference(self, other):
        """
        Returns the union of the two regions minus any areas contained in the
        intersection of the two regions.

        Parameters
        ----------
        other : Region
            Other region

        Returns
        -------
        Region
        """
        raise NotImplementedError

    @abstractmethod
    def union(self, other):
        """
        Returns a region representing the union of this region with ``other``.

        Parameters
        ----------
        other : Region
            Other region

        Returns
        -------
        Region
        """
        raise NotImplementedError

    def __and__(self, other):
        return self.intersection(other)

    def __or__(self, other):
        return self.union(other)

    def __xor__(self, other):
        return self.symmetric_difference(other)

    @abstractmethod
    def contains(self, points):
        """
        Return list of indices for all points that are inside the region of interest.

        Parameters
        ----------
        points : array-like
            Coordinates of points that are tested for being inside the specified region.

        Returns
        -------
        numpy.ndarray of ints
            Array with indices for all points in original point array that are within the region.
        """
        pass

    def __contains__(self, item):
        return True if list(self.contains([item])) == [0] else False

    @abstractmethod
    def buffer(self, distance):
        """
        Extend the region perpendicular by a `distance`.

        Parameters
        ----------
        distance : float
            Distance by which the region is extended.

        Returns
        -------
        Polygon
            The extended region.
        """
        pass


class Region1D(Region):
    """
       Abstract Region class to define the interface for 1-dimensional Region classes.
    """
    @property
    def dimension(self):
        return 1

    @abstractmethod
    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Matplotlib patch object for this region (e.g. `matplotlib.patches.Ellipse`).

        Parameters
        ----------
        origin : array_like
            The (x, y) pixel position of the origin of the displayed image.
            Default is (0, 0).
        kwargs : dict
            Other parameters passed to the `matplotlib.patches` object.

        Returns
        -------
        patch : matplotlib.patches
            Matplotlib patch for the specified region.
        """
        pass

    def intersection(self, other):
        raise NotImplementedError

    def symmetric_difference(self, other):
        raise NotImplementedError

    def union(self, other):
        raise NotImplementedError


class Region2D(Region):
    """
       Abstract Region class to define the interface for 2-dimensional Region classes.
    """
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_shapely_object']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.__dict__['_shapely_object'] = None

    @property
    def dimension(self):
        return 2

    @property
    def bounds(self):
        min_x, min_y, max_x, max_y = self.shapely_object.bounds
        return min_x, min_y, max_x, max_y

    @property
    def extent(self):
        min_x, min_y, max_x, max_y = self.bounds
        return abs(max_x - min_x), abs(max_y - min_y)

    @property
    def bounding_box(self):
        min_x, min_y, max_x, max_y = self.bounds
        return Rectangle((min_x, min_y), max_x - min_x, max_y - min_y, 0)

    @property
    @abstractmethod
    def shapely_object(self):
        """
        Geometric object as defined in `shapely`.

        Returns
        -------
        Shapely object
        """
        pass

    @staticmethod
    def from_shapely(shapely_object):
        """
        Constructor for instantiating Region from `shapely` object.

        Parameters
        ----------
        shapely_object `shapely`
            Geometric object to be converted into Region

        Returns
        -------
        Region
        """
        ptype = shapely_object.geom_type
        if ptype == 'Polygon':
            return Polygon.from_shapely(shapely_object)
        elif ptype == 'MultiPolygon':
            return MultiPolygon.from_shapely(shapely_object)

    @abstractmethod
    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Matplotlib patch object for this region (e.g. `matplotlib.patches.Ellipse`).

        Parameters
        ----------
        origin : array_like
            The (x, y) pixel position of the origin of the displayed image.
            Default is (0, 0).
        kwargs : dict
            Other parameters passed to the `matplotlib.patches` object.

        Returns
        -------
        patch : matplotlib.patches
            Matplotlib patch for the specified region.
        """
        pass

    def plot(self, ax=None, **kwargs):
        """
        Provide plot of region as :class:`matplotlib.axes.Axes` object.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        kwargs : dict
            Other parameters passed to the `matplotlib.patches` object.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        artist = self.as_artist(**kwargs)
        ax.add_artist(artist)

        return ax

    def intersection(self, other):
        shapely_obj = self.shapely_object.intersection(other.shapely_object)
        return Region2D.from_shapely(shapely_obj)

    def symmetric_difference(self, other):
        shapely_obj = self.shapely_object.symmetric_difference(other.shapely_object)
        return Region2D.from_shapely(shapely_obj)

    def union(self, other):
        shapely_obj = self.shapely_object.union(other.shapely_object)
        return Region2D.from_shapely(shapely_obj)

    def buffer(self, distance, **kwargs):
        """
        Extend the region perpendicular by a `distance`.

        Parameters
        ----------
        distance : float
            Distance by which the region is extended.
        kwargs : dict
            Other parameters passed to :func:`shapely.geometry.buffer`.

        Returns
        -------
        Polygon
            The extended region.
        """
        return Region2D.from_shapely(self.shapely_object.buffer(distance, **kwargs))


class Region3D(Region):
    """
       Abstract Region class to define the interface for 3-dimensional Region classes.
    """
    @property
    def dimension(self):
        return 3

    @abstractmethod
    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Matplotlib patch object for this region (e.g. `matplotlib.patches.Ellipse`).

        Parameters
        ----------
        origin : array_like
            The (x, y) pixel position of the origin of the displayed image.
            Default is (0, 0).
        kwargs : dict
            Other parameters passed to the `matplotlib.patches` object.

        Returns
        -------
        patch : matplotlib.patches
            Matplotlib patch for the specified region.
        """
        pass

    def plot(self, ax=None, **kwargs):
        """
        Provide plot of region as :class:`matplotlib.axes.Axes` object.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        kwargs : dict
            Other parameters passed to the `matplotlib.patches` object.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        artist = self.as_artist(**kwargs)
        ax.add_artist(artist)

        return ax

    def intersection(self, other):
        raise NotImplementedError

    def symmetric_difference(self, other):
        raise NotImplementedError

    def union(self, other):
        raise NotImplementedError


class RegionND(Region):
    """
       Abstract Region class to define the interface for n-dimensional Region classes.
    """
    def as_artist(self):
        raise NotImplementedError

    def intersection(self, other):
        raise NotImplementedError

    def symmetric_difference(self, other):
        raise NotImplementedError

    def union(self, other):
        raise NotImplementedError


class EmptyRegion(Region):
    """
       Region class to define an empty region that has no dimension.
    """
    def __init__(self):
        self.shapely_object = shPolygon()

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'{self.__class__.__name__}()'

    @property
    def dimension(self):
        return None

    @property
    def points(self):
        return np.array([])

    @property
    def centroid(self):
        return None

    @property
    def max_distance(self):
        return 0

    @property
    def region_measure(self):
        return 0

    @property
    def subregion_measure(self):
        return 0

    @property
    def bounds(self):
        return None

    @property
    def extent(self):
        return None

    @property
    def bounding_box(self):
        return None

    def intersection(self, other):
        return EmptyRegion()

    def symmetric_difference(self, other):
        return other

    def union(self, other):
        return other

    def contains(self, points):
        return np.array([])

    def as_artist(self, **kwargs):
        raise NotImplementedError("EmptyRegion cannot return an artist.")

    def buffer(self, distance, **kwargs):
        raise NotImplementedError("EmptyRegion cannot be extended.")

    @classmethod
    def from_shapely(cls, polygon):
        if polygon.is_empty:
            return cls()
        else:
            raise TypeError("Shapely object must be an empty.")


class Interval(Region1D):
    """
    Region class to define an interval.

    Parameters
    ----------
    lower_bound : float
        The lower bound of the interval.
    upper_bound : float
        The upper bound of the interval.
    """

    def __init__(self, lower_bound=0, upper_bound=1):
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self._region_specs = (self.lower_bound, self.upper_bound)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.lower_bound}, {self.upper_bound})'

    @classmethod
    def from_intervals(cls, intervals):
        """
        Constructor for instantiating Region from list of (min, max) bounds.
        Takes array-like intervals instead of interval to be consistent with `Rectangle.from_intervals`.

        Parameters
        ----------
        intervals : array-like of shape (2,)
            The region bounds for each dimension

        Returns
        -------
        Interval
        """
        lower_bound, upper_bound = intervals
        return cls(lower_bound, upper_bound)

    @property
    def lower_bound(self):
        """
        The lower boundary.

        Returns
        -------
        float
        """
        return self._lower_bound

    @property
    def upper_bound(self):
        """
        The upper boundary.

        Returns
        -------
        float
        """
        return self._upper_bound

    @property
    def bounds(self):
        return self.lower_bound, self.upper_bound

    @property
    def extent(self):
        return abs(self.upper_bound - self.lower_bound)

    @property
    def intervals(self):
        """
        Provide bounds in a tuple (min, max) arrangement.

        Returns
        -------
        Tuple of shape(dimension, 2)
            ((min_x, max_x), ...).
        """
        return (self.bounds,)

    @property
    def region_specs(self):
        """
        Legacy interface to serve legacy RoiRegion.

        Warnings
        --------
        Do not use - will be deprecated.

        Returns
        -------
        dict
        """
        return self._region_specs

    @property
    def points(self):
        return np.array([self.lower_bound, self.upper_bound])

    @property
    def centroid(self):
        return self.lower_bound + (self.upper_bound - self.lower_bound) / 2

    @property
    def max_distance(self):
        return self.upper_bound - self.lower_bound

    @property
    def region_measure(self):
        return self.upper_bound - self.lower_bound

    @property
    def subregion_measure(self):
        return None

    @property
    def bounding_box(self):
        return self

    def contains(self, points):
        points_ = np.asarray(points)
        condition = (points_ >= self.lower_bound) & (points_ < self.upper_bound)
        inside_indices = condition.nonzero()[0]  # points are 1-dimensional
        return inside_indices

    def as_artist(self, origin=(0, 0), **kwargs):
        raise NotImplementedError

    def buffer(self, distance, **kwargs):
        l_bound = self.lower_bound - distance
        u_bound = self.upper_bound + distance
        return Interval(lower_bound=l_bound, upper_bound=u_bound)


class Rectangle(Region2D):
    """
    Region class to define a rectangle.

    Parameters
    ----------
    corner : array-like with shape (2,)
        A point that defines the lower left corner.
    width : float
        The length of a vector describing the edge in x-direction.
    height : float
        The length of a vector describing the edge in y-direction.
    angle : float
        The angle (in degrees) by which the rectangle is rotated counterclockwise around the corner point.
    """

    def __init__(self, corner=(0, 0), width=1, height=1, angle=0):
        self._corner = corner
        self._width = width
        self._height = height
        self._angle = angle
        self._region_specs = (corner, width, height, angle)
        self._shapely_object = None

    def __repr__(self):
        return f'{self.__class__.__name__}({tuple(self.corner)}, {self.width}, {self.height}, {self.angle})'

    def __getattr__(self, attr):
        """All non-adapted calls are passed to shapely object"""
        if attr.startswith('__') and attr.endswith('__'):  # this is needed to enable pickling
            raise AttributeError
        return getattr(self.shapely_object, attr)

    @classmethod
    def from_intervals(cls, intervals):
        """
        Constructor for instantiating Region from list of (min, max) bounds.

        Parameters
        ----------
        intervals : tuple, list, np.ndarray of shape (2, 2)
            The region bounds for each dimension

        Returns
        -------
            Interval
        """
        min_x, max_x = intervals[0]
        min_y, max_y = intervals[1]
        corner = (min_x, min_y)
        width = max_x - min_x
        height = max_y - min_y
        angle = 0
        return cls(corner, width, height, angle)

    @property
    def corner(self):
        """
        A point that defines the lower left corner.

        Returns
        -------
        array-like with shape (2,)
        """
        return self._corner

    @property
    def width(self):
        """
        The length of a vector describing the edge in x-direction.

        Returns
        -------
        float
        """
        return self._width

    @property
    def height(self):
        """
        The length of a vector describing the edge in y-direction.

        Returns
        -------
        float
        """
        return self._height

    @property
    def angle(self):
        """
        The angle (in degrees) by which the rectangle is rotated counterclockwise around the corner point.

        Returns
        -------
        float
        """
        return self._angle

    @property
    def intervals(self):
        """
        Provide bounds in a tuple (min, max) arrangement.

        Returns
        -------
        Tuple of shape(dimension, 2)
            ((min_x, max_x), ...).
        """
        lower_bounds = self.bounds[:self.dimension]
        upper_bounds = self.bounds[self.dimension:]
        return tuple(((lower, upper) for lower, upper in zip(lower_bounds, upper_bounds)))

    @property
    def points(self):
        rectangle = mpl_patches.Rectangle(self.corner, self.width, self.height, angle=self.angle,
                                          fill=False, edgecolor='b', linewidth=1)
        points = rectangle.get_verts()
        return points[::-1]

    @property
    def region_specs(self):
        """
        Legacy interface to serve legacy RoiRegion.

        Warnings
        --------
        Do not use - will be deprecated.

        Returns
        -------
        dict
        """
        return self._region_specs

    @property
    def shapely_object(self):
        if self._shapely_object is None:
            self._shapely_object = shPolygon(self.points[:-1])
        return self._shapely_object

    @property
    def centroid(self):
        return list(self.shapely_object.centroid.coords)[0]

    @property
    def max_distance(self):
        return np.sqrt(self.width**2 + self.height**2)

    @property
    def region_measure(self):
        return self.width * self.height

    @property
    def subregion_measure(self):
        return 2 * self.width + 2 * self.height

    def contains(self, points):
        if np.asarray(points).size == 0:
            return np.array([])
        polygon_path = mpl_path.Path(self.points, closed=True)
        mask = polygon_path.contains_points(points)
        inside_indices = np.nonzero(mask)[0]
        return inside_indices

    def as_artist(self, origin=(0, 0), **kwargs):
        xy = self.corner[0] - origin[0], self.corner[1] - origin[1]
        return mpl_patches.Rectangle(xy=xy, width=self.width, height=self.height, angle=self.angle, **kwargs)


class Ellipse(Region2D):
    """
    Region class to define an ellipse.

    Parameters
    ----------
    center : array-like with shape (2,)
        A point that defines the center of the ellipse.
    width : float
        The length of a vector describing the principal axis in x-direction (before rotation).
    height : float
        The length of a vector describing the principal axis in y-direction (before rotation).
    angle : float
        The angle (in degrees) by which the ellipse is rotated counterclockwise around the center point.
    """
    def __init__(self, center=(0, 0), width=1, height=1, angle=0):
        self._center = center
        self._width = width
        self._height = height
        self._angle = angle
        self._region_specs = (center, width, height, angle)
        self._shapely_object = None

    def __repr__(self):
        return f'{self.__class__.__name__}({tuple(self.center)}, {self.width}, {self.height}, {self.angle})'

    def __getattr__(self, attr):
        """All non-adapted calls are passed to shapely object"""
        if attr.startswith('__') and attr.endswith('__'):  # this is needed to enable pickling
            raise AttributeError
        return getattr(self.shapely_object, attr)

    @property
    def center(self):
        """
        A point that defines the center of the ellipse.

        Returns
        -------
        array-like with shape (2,)
        """
        return self._center

    @property
    def width(self):
        """
        The length of a vector describing the principal axis in x-direction (before rotation).

        Returns
        -------
        float
        """
        return self._width

    @property
    def height(self):
        """
        The length of a vector describing the principal axis in y-direction (before rotation).

        Returns
        -------
        float
        """
        return self._height

    @property
    def angle(self):
        """
        The angle (in degrees) by which the ellipse is rotated counterclockwise around the center point.

        Returns
        -------
        float
        """
        return self._angle

    @property
    def points(self):
        return np.array(self.shapely_object.exterior.coords)[::-1]

    @property
    def region_specs(self):
        """
        Legacy interface to serve legacy RoiRegion.

        Warnings
        --------
        Do not use - will be deprecated.

        Returns
        -------
        dict
        """
        return self._region_specs

    @property
    def shapely_object(self):
        if self._shapely_object is None:
            circle = shPoint((0, 0)).buffer(1)
            ellipse = scale(circle, self.width / 2, self.height / 2)
            rotated_ellipse = rotate(ellipse, self.angle)
            self._shapely_object = translate(rotated_ellipse, *self.center)
        return self._shapely_object

    @property
    def centroid(self):
        return self.center

    @property
    def max_distance(self):
        return np.max([self.width, self.height])

    @property
    def region_measure(self):
        return np.pi * self.width/2 * self.height/2

    @property
    def subregion_measure(self):
        # using Ramanujan approximation
        a, b = self.width / 2, self.height / 2
        t = ((a - b) / (a + b)) ** 2
        circumference = np.pi * (a + b) * (1 + 3 * t / (10 + np.sqrt(4 - 3 * t)))
        return circumference

    def contains(self, points):
        points_ = np.asarray(points)
        if points_.size == 0:
            return np.array([])

        cos_angle = np.cos(np.radians(-self.angle))
        sin_angle = np.sin(np.radians(-self.angle))

        xc = points_[:, 0] - self.center[0]
        yc = points_[:, 1] - self.center[1]

        xct = xc * cos_angle - yc * sin_angle
        yct = xc * sin_angle + yc * cos_angle

        rad_cc = (xct ** 2 / (self.width / 2.) ** 2) + (yct ** 2 / (self.height / 2.) ** 2)

        inside_indices = np.nonzero(rad_cc < 1.)[0]
        return inside_indices

    def as_artist(self, origin=(0, 0), **kwargs):
        xy = self.center[0] - origin[0], self.center[1] - origin[1]
        return mpl_patches.Ellipse(xy=xy, width=self.width, height=self.height, angle=self.angle, **kwargs)


class Polygon(Region2D):
    """
    Region class to define a polygon.

    The polygon is constructed from a list of points that can be
    closed (i.e. the first and last point are identical) or
    not (in this case the list of points will be closed implicitly).

    Parameters
    ----------
    points  : array-like with shape (n_points, 2)
        Points that define the exterior boundary of a polygon.
    holes  : array-like with shape (n_holes, n_points, 2)
        Points that define holes within the polygon.
    """

    def __init__(self, points=((0, 0), (0, 1), (1, 1), (1, 0)), holes=None):
        if np.all(points[0] == points[-1]):
            self._points = np.array(points)
        else:
            self._points = np.append(np.array(points), [points[0]], axis=0)
        if holes is None:
            self._holes = None
        else:
            self._holes = [np.array(hole) for hole in holes]
        self._region_specs = None
        self._shapely_object = None

    def __repr__(self):
        if self.holes is None:
            return f'{self.__class__.__name__}({self.points.tolist()})'
        else:
            return f'{self.__class__.__name__}({self.points.tolist()}, {[hole.tolist() for hole in self.holes]})'

    def __str__(self):
        return f'{self.__class__.__name__}(<self.points>, <self.holes>)'

    def __getattr__(self, attr):
        """All non-adapted calls are passed to shapely object"""
        if attr.startswith('__') and attr.endswith('__'):  # this is needed to enable pickling
            raise AttributeError
        return getattr(self.shapely_object, attr)

    @classmethod
    def from_shapely(cls, polygon):
        if polygon.is_empty:
            return EmptyRegion()
        else:
            points = np.array(polygon.exterior.coords).tolist()
            holes = [np.array(interiors.coords).tolist() for interiors in polygon.interiors]
            return cls(points, holes)

    @property
    def points(self):
        """
        Exterior polygon points.

        Returns
        -------
        numpy.array of shape(n_points, dimension)
        """
        return self._points

    @property
    def holes(self):
        """
        Holes where each hole is specified by polygon points.

        Returns
        -------
        list of numpy.array of shape(n_holes, n_points, dimension)
        """
        return self._holes

    @property
    def region_specs(self):
        """
        Legacy interface to serve legacy RoiRegion.

        Warnings
        --------
        Do not use - will be deprecated.

        Returns
        -------
        dict
        """
        return self._region_specs

    @property
    def shapely_object(self):
        if self._shapely_object is None:
            self._shapely_object = shPolygon(self.points, self.holes)
        return self._shapely_object

    @property
    def centroid(self):
        return list(self.shapely_object.centroid.coords)[0]

    @property
    def max_distance(self):
        distances = pdist(self.points[:-1])
        return np.nanmax(distances)

    @property
    def region_measure(self):
        return self.shapely_object.area

    @property
    def subregion_measure(self):
        return self.shapely_object.length

    def contains(self, points):
        _points = np.asarray(points)
        if _points.size == 0:
            return np.array([], dtype=bool)

        # preselect points inside the polygons bounding box to increase performance.
        preselected_points_indices = self.bounding_box.contains(_points)
        if len(preselected_points_indices) == 0:
            return np.array([], dtype=bool)

        points_ = shMultiPoint(_points[preselected_points_indices])
        prepared_polygon = prep(self.shapely_object)
        mask = list(map(prepared_polygon.contains, points_.geoms))
        inside_indices = np.nonzero(mask)[0]
        if len(inside_indices) == 0:
            return np.array([], dtype=bool)
        else:
            return preselected_points_indices[inside_indices]

    def as_artist(self, **kwargs):
        return mpl_patches.PathPatch(_polygon_path(self.shapely_object), **kwargs)


class MultiPolygon(Region2D):
    """
    Region class to define a region that represents the union of multiple polygons.

    Parameters
    ----------
    polygons  : list of Polygon
        Polygons that define the individual polygons.
    """
    def __init__(self, polygons):
        self._polygons = polygons
        self._region_specs = None
        self._shapely_object = None

    def __repr__(self):
        return f'{self.__class__.__name__}({self.polygons})'

    def __str__(self):
        return f'{self.__class__.__name__}(<self.polygons>)'

    def __getattr__(self, attr):
        """All non-adapted calls are passed to shapely object"""
        if attr.startswith('__') and attr.endswith('__'):  # this is needed to enable pickling
            raise AttributeError
        return getattr(self.shapely_object, attr)

    @property
    def points(self):
        """
        Exterior polygon points.

        Returns
        -------
        list of numpy.array of shape(n_polygons, n_points, dimension)
        """
        return [pol.points for pol in self.polygons]

    @property
    def holes(self):
        """
        Points defining holes.

        Returns
        -------
        list
            list of polygon holes
        """
        return [pol.holes for pol in self.polygons]

    @property
    def polygons(self):
        """
        All polygons that make up the MultiPolygon

        Returns
        -------
        list of Polygon
        """
        return self._polygons

    @property
    def region_specs(self):
        """
        Legacy interface to serve legacy RoiRegion.

        Warnings
        --------
        Do not use - will be deprecated.

        Returns
        -------
        dict
        """
        return self._region_specs

    @property
    def shapely_object(self):
        if self._shapely_object is None:
            self._shapely_object = shMultiPolygon([pol.shapely_object for pol in self._polygons])
        return self._shapely_object

    @property
    def centroid(self):
        return list(self.shapely_object.centroid.coords)[0]

    @property
    def max_distance(self):
        distances = pdist(np.array([point for pts in self.points for point in pts]))
        return np.nanmax(distances)

    @property
    def region_measure(self):
        return self.shapely_object.area

    @property
    def subregion_measure(self):
        return self.shapely_object.length

    def contains(self, points):
        if np.asarray(points).size == 0:
            return np.array([])
        points_ = shMultiPoint(points)
        prepared_polygon = prep(self.shapely_object)
        mask = list(map(prepared_polygon.contains, points_.geoms))
        inside_indices = np.nonzero(mask)[0]
        return inside_indices

    def as_artist(self, **kwargs):
        polygon = shMultiPolygon([pol.shapely_object for pol in self.polygons])
        return mpl_patches.PathPatch(_polygon_path(polygon), **kwargs)

    @classmethod
    def from_shapely(cls, multipolygon):
        return cls([Polygon.from_shapely(pol) for pol in multipolygon.geoms])


class AxisOrientedCuboid(Region3D):
    """
    Region class to define an axis-oriented cuboid.

    This is a 3-dimensional convex region with rectangular faces
    and edges that are parallel to coordinate axes.
    Extension in x-, y-, z-coordinates correspond to length, width, height.

    Parameters
    ----------
    corner : array-like with shape (3,)
        A point that defines the lower left corner.
    length : float
        The length of a vector describing the edge in x-direction.
    width : float
        The length of a vector describing the edge in y-direction.
    height : float
        The length of a vector describing the edge in z-direction.
    """
    def __init__(self, corner=(0, 0, 0), length=1, width=1, height=1):
        self._corner = corner
        self._length = length
        self._width = width
        self._height = height

    def __repr__(self):
        return f'{self.__class__.__name__}({tuple(self.corner)}, ' \
               f'{self.length}, {self.width}, {self.height})'

    @classmethod
    def from_intervals(cls, intervals):
        """
        Constructor for instantiating Region from list of (min, max) bounds.

        Parameters
        ----------
        intervals : array-like of shape (3, 2)
            The region bounds for each dimension

        Returns
        -------
            Interval
        """
        min_x, max_x = intervals[0]
        min_y, max_y = intervals[1]
        min_z, max_z = intervals[2]
        corner = (min_x, min_y, min_z)
        length = max_x - min_x
        width = max_y - min_y
        height = max_z - min_z
        return cls(corner, length, width, height)

    @property
    def corner(self):
        """
        A point that defines the lower left corner.

        Returns
        -------
        array-like with shape (2,)
        """
        return self._corner

    @property
    def length(self):
        """
        The length of a vector describing the edge in x-direction.

        Returns
        -------
        float
        """
        return self._length

    @property
    def width(self):
        """
        The length of a vector describing the edge in y-direction.

        Returns
        -------
        float
        """
        return self._width

    @property
    def height(self):
        """
        The length of a vector describing the edge in z-direction.

        Returns
        -------
        float
        """
        return self._height

    @property
    def points(self):
        return tuple(it.product(*self.intervals))

    @property
    def centroid(self):
        return tuple((cor + dist / 2 for cor, dist in zip(self.corner, (self.length, self.width, self.height))))

    @property
    def bounds(self):
        min_x, min_y, min_z = self.corner
        max_x, max_y, max_z = [cor + dist for cor, dist in zip(self.corner, (self.length, self.width, self.height))]
        return min_x, min_y, min_z, max_x, max_y, max_z

    @property
    def intervals(self):
        """
        Provide bounds in a tuple (min, max) arrangement.

        Returns
        -------
        Tuple of shape(dimension, 2)
            ((min_x, max_x), ...).
        """
        min_x, min_y, min_z, max_x, max_y, max_z = self.bounds
        return ((min_x, max_x), (min_y, max_y), (min_z, max_z))

    @property
    def extent(self):
        return abs(self.length), abs(self.width), abs(self.height)

    @property
    def bounding_box(self):
        return self

    @property
    def max_distance(self):
        return np.sqrt(self.length**2 + self.width**2 + self.height**2)

    @property
    def region_measure(self):
        return self.length * self.width * self.height

    @property
    def subregion_measure(self):
        return 2 * (self.length * self.width + self.height * self.width + self.height * self.length)

    def contains(self, points):
        points = np.asarray(points)
        if points.size == 0:
            return np.array([])
        condition_0 = [(points[:, i] >= bound) for i, bound in enumerate(self.bounds[:3])]
        condition_1 = [(points[:, i] < bound) for i, bound in enumerate(self.bounds[3:])]
        condition = np.all(condition_0, axis=0) & np.all(condition_1, axis=0)
        inside_indices = np.nonzero(condition)[0]
        return inside_indices

    def as_artist(self, origin=(0, 0), **kwargs):
        raise NotImplementedError

    def buffer(self, distance, **kwargs):
        mins = self.bounds[:self.dimension]
        maxs = self.bounds[self.dimension:]
        new_mins = [value - distance for value in mins]
        new_maxs = [value + distance for value in maxs]
        return AxisOrientedCuboid(new_mins, *(max_ - min_ for max_, min_ in zip(new_maxs, new_mins)))


# todo: complete implementation
class Cuboid(Region3D):
    """
    Region class to define a cuboid.

    This is a 3-dimensional convex region with rectangular faces.
    Extension in x-, y-, z-coordinates correspond to length, width, height.
    Corresponding Euler angles are defined by alpha, beta, gamma.

    Parameters
    ----------
    corner : array-like with shape (2,)
        A point that defines the lower left corner.
    length : float
        The length of a vector describing the edge in x-direction.
    width : float
        The length of a vector describing the edge in y-direction.
    alpha : float
        The first Euler angle (in degrees) by which the cuboid is rotated.
    beta : float
        The second Euler angle (in degrees) by which the cuboid is rotated.
    gamma : float
        The third Euler angle (in degrees) by which the cuboid is rotated.
    """
    def __init__(self, corner=(0, 0, 0), length=1, width=1, height=1, alpha=0, beta=0, gamma=0):
        self._corner = corner
        self._length = length
        self._width = width
        self._height = height
        self._alpha = alpha
        self._beta = beta
        self._gamma = gamma
        self._region_specs = (corner, length, width, height, alpha, beta, gamma)

    def __repr__(self):
        return f'{self.__class__.__name__}({tuple(self.corner)}, ' \
               f'{self.length}, {self.width}, {self.height}, ' \
               f'{self.alpha}, {self.beta}, {self.gamma})'

    @property
    def corner(self):
        """
        A point that defines the lower left corner.

        Returns
        -------
        array-like with shape (2,)
        """
        return self._corner

    @property
    def length(self):
        """
        The length of a vector describing the edge in x-direction.

        Returns
        -------
        float
        """
        return self._length

    @property
    def width(self):
        """
        The length of a vector describing the edge in y-direction.

        Returns
        -------
        float
        """
        return self._width

    @property
    def height(self):
        """
        The length of a vector describing the edge in z-direction.

        Returns
        -------
        float
        """
        return self._height

    @property
    def alpha(self):
        """
        The first Euler angle (in degrees) by which the cuboid is rotated.

        Returns
        -------
        float
        """
        return self._alpha

    @property
    def beta(self):
        """
        The sescond Euler angle (in degrees) by which the cuboid is rotated.

        Returns
        -------
        float
        """
        return self._beta

    @property
    def gamma(self):
        """
        The third Euler angle (in degrees) by which the cuboid is rotated.

        Returns
        -------
        float
        """
        return self._gamma

    @property
    def points(self):
        raise NotImplementedError

    @property
    def bounds(self):
        raise NotImplementedError

    @property
    def extent(self):
        raise NotImplementedError

    @property
    def centroid(self):
        raise NotImplementedError

    @property
    def max_distance(self):
        return np.sqrt(self.length**2 + self.width**2 + self.height**2)

    @property
    def region_measure(self):
        return self.length * self.width * self.height

    @property
    def subregion_measure(self):
        return 2 * (self.length * self.width + self.height * self.width + self.height * self.length)

    def contains(self, points):
        raise NotImplementedError

    def as_artist(self, origin=(0, 0), **kwargs):
        raise NotImplementedError

    def buffer(self, distance, **kwargs):
        raise NotImplementedError

    @property
    def bounding_box(self):
        return self

class AxisOrientedHypercuboid(RegionND):
    """
    Region class to define an axis-oriented n-dimensional hypercuboid.

    This is a n-dimensional convex region with rectangular faces
    and edges that are parallel to coordinate axes.
    Extension in x-, y-, z-coordinates correspond to length, width, height.

    Parameters
    ----------
    corner : array-like with shape (dimension,)
        A point that defines the lower left corner.
    lengths : array-like of shape(dimension,)
        Array of length values for the 1-dimensional edge vectors.
    """
    def __init__(self, corner=(0, 0, 0), lengths=(1, 1, 1)):
        if not len(corner) == len(lengths):
            raise TypeError("corner and lengths must have the same dimension.")
        self._corner = np.array(corner)
        self._lengths = np.array(lengths)

    def __repr__(self):
        return f'{self.__class__.__name__}({tuple(self.corner)}, {tuple(self.lengths)})'

    @classmethod
    def from_intervals(cls, intervals):
        """
        Constructor for instantiating Region from list of (min, max) bounds.

        Parameters
        ----------
        intervals : array-like of shape (dimension, 2)
            The region bounds for each dimension

        Returns
        -------
            Interval
        """
        intervals = np.array(intervals)
        corner = intervals[:, 0]
        lengths = np.diff(intervals)[:, 0]
        return cls(corner, lengths)

    @property
    def corner(self):
        """
        A point that defines the lower left corner.

        Returns
        -------
        array-like with shape (dimension,)
        """
        return self._corner

    @property
    def lengths(self):
        """
        Array of length values for the 1-dimensional edge vectors.

        Returns
        -------
        array-like of shape(dimension,)
        """
        return self._lengths

    @property
    def dimension(self):
        return len(self.lengths)

    @property
    def intervals(self):
        """
        Provide bounds in a tuple (min, max) arrangement.

        Returns
        -------
        Tuple of shape(dimension, 2)
            ((min_x, max_x), ...).
        """
        return tuple((lower, upper) for lower, upper in zip(self.bounds[:self.dimension], self.bounds[self.dimension:]))

    @property
    def points(self):
        return tuple(it.product(*self.intervals))

    @property
    def bounds(self):
        return np.concatenate([self.corner, self.corner + self.lengths], axis=0)

    @property
    def extent(self):
        return np.abs(self.lengths)

    @property
    def centroid(self):
        return self.corner + self.lengths / 2

    @property
    def max_distance(self):
        return np.sqrt(np.sum(self.lengths**2))

    @property
    def region_measure(self):
        return np.product(self.lengths)

    @property
    def subregion_measure(self):
        raise NotImplementedError

    @property
    def bounding_box(self):
        return self

    def contains(self, points):
        points = np.asarray(points)
        if points.size == 0:
            return np.array([])
        condition_0 = [(points[:, i] >= bound) for i, bound in enumerate(self.bounds[:self.dimension])]
        condition_1 = [(points[:, i] < bound) for i, bound in enumerate(self.bounds[self.dimension:])]
        condition = np.all(condition_0, axis=0) & np.all(condition_1, axis=0)
        inside_indices = np.nonzero(condition)[0]
        return inside_indices

    def buffer(self, distance):
        mins = self.bounds[:self.dimension] - distance
        maxs = self.bounds[self.dimension:] + distance
        lengths = maxs - mins
        return AxisOrientedHypercuboid(mins, lengths)


def _polygon_path(polygon):
    """
    Constructs a compound matplotlib path from a Shapely geometric object.
    Adapted from https://pypi.org/project/descartes/ (BSD license, copyright Sean Gillies)
    """
    def coding(ob):
        # The codes will be all "LINETO" commands, except for "MOVETO"s at the
        # beginning of each subpath
        n = len(getattr(ob, 'coords', None) or ob)
        vals = np.ones(n, dtype=mpl_path.Path.code_type) * mpl_path.Path.LINETO
        vals[0] = mpl_path.Path.MOVETO
        return vals

    ptype = polygon.geom_type
    if ptype == 'Polygon':
        polygon = [polygon]
    elif ptype == 'MultiPolygon':
        polygon = [shPolygon(p) for p in polygon.geoms]
    else:
        raise ValueError(
            "A polygon or multi-polygon representation is required")

    vertices = np.concatenate([
        np.concatenate([np.asarray(t.exterior.coords)[:, :2]] +
                    [np.asarray(r.coords)[:, :2] for r in t.interiors])
        for t in polygon])
    codes = np.concatenate([
        np.concatenate([coding(t.exterior)] +
                    [coding(r) for r in t.interiors]) for t in polygon])

    return mpl_path.Path(vertices, codes)



#!/usr/bin/env python

"""
Render localization data in napari.

With this script you can choose a file name and render the localization file in napari.

To run the script::

    locan napari -f <file> -t <file type> -k <string with kwargs for render function> --bin_size <>
    --rescale <string with tuple or rescale>

Try for instance::

    locan napari -f "locan/tests/test_data/five_blobs.txt" -t 1 --bin_size 10 --rescale "0 1"

See Also
--------
locan.render.render2d.render_2d_napari
"""
import re
import argparse
from pathlib import Path

from locan.dependencies import HAS_DEPENDENCY
from locan.constants import FileType
from locan.gui.io import file_dialog
import locan.locan_io.locdata.io_locdata as io
from locan.render.render2d import render_2d_napari

if HAS_DEPENDENCY["napari"]: import napari


def sc_napari(file_path=None, file_type=FileType.CUSTOM, **kwargs):
    """
    Render localization data in napari.

    Parameters
    ----------
    file_path : str, bytes, os.PathLike
        File path to localization data.
    file_type : int, str, locan.constants.FileType, locan.data.metadata_pb2.Metadata
        Indicator for the file type.
        Integer or string should be according to locan.constants.FileType.
    kwargs : dict
        Other parameters passed to :func:`render_2d_napari`.
    """
    # choose file interactively
    if file_path is None:
        file_path = Path(file_dialog(message='choose file', filter=
                    'Text files (*.txt);; CSV files (*.csv);; All files (*);; All files ()'
                    )[0])
    else:
        file_path = Path(file_path)

    print(file_path)

    # load data
    dat = io.load_locdata(path=file_path, file_type=file_type)

    # render
    render_2d_napari(locdata=dat, **kwargs)
    napari.run()


def type_converter_rescale(input_string):
    if input_string == 'None':
        return None
    elif input_string == 'True':
        return True
    elif input_string == 'False':
        return False
    else:
        pattern = re.match(r'\(?([0-9]*[.]?[0-9]+),?\s?([0-9]*[.]?[0-9]+)\)?', input_string)
        if pattern:
            return tuple(float(element) for element in pattern.groups())
        else:
            return input_string


def _add_arguments(parser):
    parser.add_argument('-f', '--file', dest='file', type=str, default=None,
                        help='File path to localization data.')
    parser.add_argument('-t', '--type', dest='type', type=int, default=2,
                        help='Integer or string indicating the file type.')
    parser.add_argument('--bin_size', dest='bin_size', type=float, default=10,
                        help='The size of bins in units of locdata coordinate units. '
                             'Keyword passed to render_2d_napari function.')
    parser.add_argument('--rescale', dest='rescale', type=type_converter_rescale, default='EQUALIZE',
                        help='Rescale intensity values. Keyword passed to render_2d_napari function.')


def main(args=None):

    parser = argparse.ArgumentParser(description='Render localization data in napari.')
    _add_arguments(parser)
    returned_args = parser.parse_args(args)

    sc_napari(returned_args.file, returned_args.type, bin_size=returned_args.bin_size, rescale=returned_args.rescale)


if __name__ == '__main__':
    main()



"""
Miscellaneous functions without a home.
"""
from __future__ import annotations

from collections.abc import Generator


__all__ = ['iterate_2d_array']


def iterate_2d_array(n_elements=5, n_cols=2) -> Generator[tuple[int, int]]:
    """
    Iterator for 2-dimensional array iterating first over columns then over rows.

    Parameters
    ----------
    n_elements : int
        Number of elements
    n_cols : int
        Number of columns

    Returns
    -------
    Generator[tuple[int, int]]
        Indices for (row, column) in each iteration.
    """
    n_rows = -(-n_elements // n_cols)
    iterator = ((i, j) for i in range(n_rows) for j in range(n_cols))
    return (item for item, _ in zip(iterator, range(n_elements)))



"""
Command-line utility scripts

The locan package provides a command-line interface.
It is accessible by the base command ``locan`` and provides a number of compound commands.
See available commands using ``locan -h``.

This subpackage contains implementations of command-line scripts.
Scripts are installed in bin/ as simple wrappers for these modules.
They can be run directly from a terminal through ``locan command options``
as long as the correct environment is activated.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   script_check
   script_rois
   script_draw_roi
   script_napari
   script_show_versions
   script_test

"""



"""

File input/output for localization data in rapidSTORM files.

"""
import logging

import numpy as np
import pandas as pd
import xml.etree.ElementTree as etree

from locan.data.locdata import LocData
import locan.constants
from locan.data import metadata_pb2
from locan.locan_io.locdata.utilities import convert_property_types, open_path_or_file_like, convert_property_names


__all__ = ['load_rapidSTORM_header', 'load_rapidSTORM_file',
           'load_rapidSTORM_track_header', 'load_rapidSTORM_track_file']

logger = logging.getLogger(__name__)


def _read_rapidSTORM_header(file):
    """
    Read xml header from a rapidSTORM single-molecule localization file and identify column names.

    Parameters
    ----------
    file : file-like
        A rapidSTORM file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """
    # read xml part in header
    header = file.readline()
    header = header[2:]

    # get iteratible
    parsed = etree.XML(header)

    # list identifiers
    identifiers = []
    for elem in parsed:
        for name, value in sorted(elem.attrib.items()):
            if name == 'identifier':
                identifiers.append(value)

    # turn identifiers into valuable LocData keys
    column_keys = convert_property_names(properties=identifiers, property_mapping=locan.constants.RAPIDSTORM_KEYS)
    return column_keys


def load_rapidSTORM_header(path):
    """
    Load xml header from a rapidSTORM single-molecule localization file and identify column names.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a rapidSTORM file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """

    # read xml part in header
    with open_path_or_file_like(path) as file:
        return _read_rapidSTORM_header(file)


def load_rapidSTORM_file(path, nrows=None, convert=True, **kwargs):
    """
    Load data from a rapidSTORM single-molecule localization file.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a rapidSTORM file to load.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.
    convert : bool
        If True convert types by applying type specifications in locan.constants.PROPERTY_KEYS.
    kwargs : dict
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with open_path_or_file_like(path) as file:
        columns = _read_rapidSTORM_header(file)
        dataframe = pd.read_csv(file, sep=" ", skiprows=0, nrows=nrows, names=columns, **kwargs)

    if convert:
        dataframe = convert_property_types(dataframe, types=locan.constants.PROPERTY_KEYS)

    dat = LocData.from_dataframe(dataframe=dataframe)

    dat.meta.source = metadata_pb2.EXPERIMENT
    dat.meta.state = metadata_pb2.RAW
    dat.meta.file.type = metadata_pb2.RAPIDSTORM
    dat.meta.file.path = str(path)

    for property in sorted(list(set(columns).intersection({'position_x', 'position_y', 'position_z'}))):
        dat.meta.localization_properties.add(name=property, unit="nm", type="float")

    del dat.meta.history[:]
    dat.meta.history.add(name='load_rapidSTORM_file', parameter='path={}, nrows={}'.format(path, nrows))

    return dat


def _read_rapidSTORM_track_header(file):
    """
    Read xml header from a rapidSTORM (track) single-molecule localization file and identify column names.

    Parameters
    ----------
    file : file-like
        A rapidSTORM file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """
    # read xml part in header
    header = file.readline()
    header = header[2:]

    parsed = etree.XML(header)

    # list identifiers
    identifiers = []
    for field in parsed.findall('field'):
        identifiers.append(field.get('identifier'))

    # turn identifiers into valuable LocData keys
    column_keys = convert_property_names(properties=identifiers, property_mapping=locan.constants.RAPIDSTORM_KEYS)

    # list child identifiers
    child_identifiers = []
    for field in parsed.findall('localizations'):
        for field_ in field.findall('field'):
            child_identifiers.append(field_.get('identifier'))

    # turn child identifiers into valuable LocData keys
    column_keys_tracks = convert_property_names(properties=child_identifiers,
                                                property_mapping=locan.constants.RAPIDSTORM_KEYS)

    return column_keys, column_keys_tracks


def load_rapidSTORM_track_header(path):
    """
    Load xml header from a rapidSTORM (track) single-molecule localization file and identify column names.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a rapidSTORM file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """

    # read xml part in header
    with open_path_or_file_like(path) as file:
        return _read_rapidSTORM_track_header(file)


def load_rapidSTORM_track_file(path, nrows=None, convert=True, collection=True, min_localization_count=1, **kwargs):
    """
    Load data from a rapidSTORM single-molecule localization file with tracked localizations.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a rapidSTORM file to load.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.
    convert : bool
        If True convert types by applying type specifications in locan.constants.PROPERTY_KEYS.
    collection : bool
        If True a collection of all tracks is returned. If False LocData with center positions is returned.
    min_localization_count : int
        If collection is True, only clusters with at least `min_localization_count` localizations are loaded.
    kwargs : dict
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations/tracks as a collection.
    """
    with open_path_or_file_like(path) as file:
        columns, columns_track = _read_rapidSTORM_track_header(file)
        lines = pd.read_csv(file, lineterminator='\n', nrows=nrows, skiprows=1, header=None, **kwargs)

    lines = lines[0].str.split(" ", expand=False)

    if collection:
        # prepare dataframes with tracked localizations
        tracks = [np.array(line[len(columns) + 1:]).reshape(-1, len(columns_track)) for line in lines
                  if int(line[len(columns)]) >= min_localization_count]
        # +1 to account for the column with number of locs in track
        track_list = []
        for track in tracks:
            dataframe = pd.DataFrame(track, columns=columns_track)
            if convert:
                dataframe = convert_property_types(dataframe, types=locan.constants.PROPERTY_KEYS)
            else:
                dataframe = dataframe.convert_dtypes()
            dat = LocData.from_dataframe(dataframe=dataframe)
            track_list.append(dat)

        collection = LocData.from_collection(track_list)

        collection.meta.source = metadata_pb2.EXPERIMENT
        collection.meta.state = metadata_pb2.RAW
        collection.meta.file.type = metadata_pb2.RAPIDSTORMTRACK
        collection.meta.file.path = str(path)

        for property in sorted(list(set(columns_track).intersection({'position_x', 'position_y', 'position_z'}))):
            collection.meta.localization_properties.add(name=property, unit="nm", type="float")

        del collection.meta.history[:]
        collection.meta.history.add(name='load_rapidSTORM_track_file',
                                    parameter='path={}, nrows={}'.format(path, nrows))

        return collection

    else:
        # prepare dataframe with center track positions
        dataframe = pd.DataFrame([line[:len(columns)] for line in lines], columns=columns)
        if convert:
            dataframe = convert_property_types(dataframe, types=locan.constants.PROPERTY_KEYS)
        else:
            dataframe = dataframe.convert_dtypes()

        locdata = LocData.from_dataframe(dataframe=dataframe)

        locdata.meta.source = metadata_pb2.EXPERIMENT
        locdata.meta.state = metadata_pb2.RAW
        locdata.meta.file.type = metadata_pb2.RAPIDSTORM
        locdata.meta.file.path = str(path)

        for property in sorted(list(set(columns).intersection({'position_x', 'position_y', 'position_z'}))):
            locdata.meta.localization_properties.add(name=property, unit="nm", type="float")

        del locdata.meta.history[:]
        locdata.meta.history.add(name='load_rapidSTORM_track_file', parameter='path={}, nrows={}'.format(path, nrows))

        return locdata



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: locan/locan_io/locdata/manifest.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from locan.data import metadata_pb2 as locan_dot_data_dot_metadata__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%locan/locan_io/locdata/manifest.proto\x12\x16locan.locan_io.locdata\x1a\x19locan/data/metadata.proto\"\xf1\x01\n\x06\x46ormat\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\x04type\x18\x02 \x01(\x0e\x32\x1c.locan.locan_io.locdata.Type\x12*\n\x04mode\x18\x03 \x01(\x0e\x32\x1c.locan.locan_io.locdata.Mode\x12\x0f\n\x07\x63olumns\x18\x04 \x01(\x03\x12\x0f\n\x07headers\x18\x05 \x03(\t\x12,\n\x05\x64type\x18\x06 \x03(\x0e\x32\x1d.locan.locan_io.locdata.Dtype\x12\r\n\x05shape\x18\x07 \x03(\x03\x12\r\n\x05units\x18\x08 \x03(\t\x12\x13\n\x0b\x64\x65scription\x18\t \x01(\t\"\xba\x03\n\x08\x46ileInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\x04type\x18\x02 \x01(\x0e\x32\x1c.locan.locan_io.locdata.Type\x12\x0e\n\x06\x66ormat\x18\x03 \x01(\t\x12\x0f\n\x07\x63hannel\x18\x04 \x01(\t\x12\x0c\n\x04rows\x18\x05 \x01(\x03\x12<\n\x06offset\x18\x06 \x03(\x0b\x32,.locan.locan_io.locdata.FileInfo.OffsetEntry\x12\x36\n\x03min\x18\x07 \x03(\x0b\x32).locan.locan_io.locdata.FileInfo.MinEntry\x12\x36\n\x03max\x18\x08 \x03(\x0b\x32).locan.locan_io.locdata.FileInfo.MaxEntry\x12\x10\n\x08\x65xposure\x18\t \x01(\x02\x1a-\n\x0bOffsetEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a*\n\x08MinEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1a*\n\x08MaxEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\"\x98\x04\n\x08Manifest\x12\x16\n\x0e\x66ormat_version\x18\x01 \x01(\t\x12>\n\x07\x66ormats\x18\x02 \x03(\x0b\x32-.locan.locan_io.locdata.Manifest.FormatsEntry\x12/\n\x05\x66iles\x18\x03 \x03(\x0b\x32 .locan.locan_io.locdata.FileInfo\x12\x0c\n\x04name\x18\x04 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x0c\n\x04tags\x18\x06 \x03(\t\x12\x11\n\tthumbnail\x18\x07 \x01(\t\x12\x0e\n\x06sample\x18\x08 \x01(\t\x12\x10\n\x08labeling\x18\t \x01(\t\x12\x0c\n\x04\x64\x61te\x18\n \x01(\t\x12\x0e\n\x06\x61uthor\x18\x0b \x01(\t\x12\x10\n\x08\x63itation\x18\x0c \x01(\t\x12\r\n\x05\x65mail\x18\r \x01(\t\x12\x36\n\x03map\x18\x0e \x03(\x0b\x32).locan.locan_io.locdata.Manifest.MapEntry\x12*\n\x0clocdata_meta\x18\x0f \x01(\x0b\x32\x14.locan.data.Metadata\x1aN\n\x0c\x46ormatsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12-\n\x05value\x18\x02 \x01(\x0b\x32\x1e.locan.locan_io.locdata.Format:\x02\x38\x01\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01*\x1c\n\x04Type\x12\t\n\x05TABLE\x10\x00\x12\t\n\x05IMAGE\x10\x01*\x1c\n\x04Mode\x12\n\n\x06\x42INARY\x10\x00\x12\x08\n\x04TEXT\x10\x01*{\n\x05\x44type\x12\x08\n\x04INT8\x10\x01\x12\t\n\x05UINT8\x10\x02\x12\t\n\x05INT16\x10\x03\x12\n\n\x06UINT16\x10\x04\x12\t\n\x05INT32\x10\x05\x12\n\n\x06UINT32\x10\x06\x12\t\n\x05INT64\x10\x07\x12\n\n\x06UINT64\x10\x08\x12\x0b\n\x07\x46LOAT32\x10\t\x12\x0b\n\x07\x46LOAT64\x10\n')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'locan.locan_io.locdata.manifest_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _FILEINFO_OFFSETENTRY._options = None
  _FILEINFO_OFFSETENTRY._serialized_options = b'8\001'
  _FILEINFO_MINENTRY._options = None
  _FILEINFO_MINENTRY._serialized_options = b'8\001'
  _FILEINFO_MAXENTRY._options = None
  _FILEINFO_MAXENTRY._serialized_options = b'8\001'
  _MANIFEST_FORMATSENTRY._options = None
  _MANIFEST_FORMATSENTRY._serialized_options = b'8\001'
  _MANIFEST_MAPENTRY._options = None
  _MANIFEST_MAPENTRY._serialized_options = b'8\001'
  _TYPE._serialized_start=1320
  _TYPE._serialized_end=1348
  _MODE._serialized_start=1350
  _MODE._serialized_end=1378
  _DTYPE._serialized_start=1380
  _DTYPE._serialized_end=1503
  _FORMAT._serialized_start=93
  _FORMAT._serialized_end=334
  _FILEINFO._serialized_start=337
  _FILEINFO._serialized_end=779
  _FILEINFO_OFFSETENTRY._serialized_start=646
  _FILEINFO_OFFSETENTRY._serialized_end=691
  _FILEINFO_MINENTRY._serialized_start=693
  _FILEINFO_MINENTRY._serialized_end=735
  _FILEINFO_MAXENTRY._serialized_start=737
  _FILEINFO_MAXENTRY._serialized_end=779
  _MANIFEST._serialized_start=782
  _MANIFEST._serialized_end=1318
  _MANIFEST_FORMATSENTRY._serialized_start=1196
  _MANIFEST_FORMATSENTRY._serialized_end=1274
  _MANIFEST_MAPENTRY._serialized_start=1276
  _MANIFEST_MAPENTRY._serialized_end=1318
# @@protoc_insertion_point(module_scope)



import pytest
import pandas as pd
from scipy import stats

from locan import LocData
from locan.analysis import LocalizationProperty
from locan.analysis.localization_property import _DistributionFits


def test_test_LocalizationProperty_empty(caplog):
    lprop = LocalizationProperty().compute(LocData())
    lprop.fit_distributions()
    lprop.plot()
    lprop.hist()
    assert caplog.record_tuples == [('locan.analysis.localization_property', 30, 'Locdata is empty.'),
                                    ('locan.analysis.localization_property', 30, 'No results available to fit.')]


def test_DistributionFits(locdata_rapidSTORM_2d):
    lprop = LocalizationProperty().compute(locdata=locdata_rapidSTORM_2d)
    assert repr(lprop) == "LocalizationProperty(loc_property=intensity, index=None)"
    ds = _DistributionFits(lprop)
    ds.fit(distribution=stats.expon)
    # print(ds.parameters)
    # print(ds.__dict__)
    # #print(ds.parameter_dict())
    # ds.plot()
    assert(ds.parameters == ['intensity_loc', 'intensity_scale'])
    ds.fit(distribution=stats.expon, with_constraints=False, floc=1000)
    assert(ds.parameter_dict()['intensity_loc'] == 1000)


def test_Localization_property(locdata_rapidSTORM_2d):
    lprop = LocalizationProperty(loc_property='intensity')
    assert not lprop
    lprop.plot()
    lprop.hist()
    lprop.compute(locdata=locdata_rapidSTORM_2d)
    assert lprop
    # print(lprop)
    assert repr(lprop) == "LocalizationProperty(loc_property=intensity, index=None)"
    assert(lprop.results.columns == pd.Index(['intensity'], dtype='object'))
    assert(lprop.distribution_statistics is None)
    lprop.plot()
    lprop.hist()
    lprop.fit_distributions()
    assert(list(lprop.distribution_statistics.parameter_dict().keys()) == ['intensity_loc', 'intensity_scale'])

    lprop = LocalizationProperty(loc_property='intensity', index='frame').compute(locdata=locdata_rapidSTORM_2d)
    # print(lprop.results)
    assert(lprop.results.index.name == 'frame')
    assert(lprop.results.columns == pd.Index(['intensity'], dtype='object'))



import pytest
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt  # needed for visual inspection

from locan import LocData, EmptyRegion, Interval, Rectangle, Ellipse, Polygon
from locan.simulation import make_uniform, make_Poisson, \
    make_cluster, make_NeymanScott, make_Matern, make_Thomas, make_dstorm
from locan.simulation import simulate_uniform, simulate_Poisson, \
    simulate_cluster, simulate_NeymanScott, simulate_Matern, simulate_Thomas, simulate_dstorm

from locan import resample, simulate_tracks, add_drift, simulate_frame_numbers
from locan.simulation.simulate_drift import _random_walk_drift, _drift


def test_make_uniform():
    rng = np.random.default_rng(seed=1)
    samples = make_uniform(n_samples=10, region=EmptyRegion(), seed=rng)
    assert np.size(samples) == 0
    samples = make_uniform(n_samples=10, region=(0, 1), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    samples = make_uniform(n_samples=10, region=((0, 1), (10, 11)), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples = make_uniform(n_samples=10, region=((0, 1), (10, 11), (100, 101)), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 3
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples = make_uniform(n_samples=10, region=Interval(0, 1), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    samples = make_uniform(n_samples=10, region=Rectangle((0, 10), 1, 1, 0), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples = make_uniform(n_samples=10, region=Ellipse((0, 0), 1, 1, 0), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 2
    assert np.all(-0.5 <= samples[:, 0])
    assert np.all(samples[:, 0] < 0.5)
    assert np.all(-0.5 <= samples[:, 1])
    assert np.all(samples[:, 1] < 0.5)

    samples = make_uniform(n_samples=10, region=Ellipse((0, 0), 1, 2, 0), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 2
    assert np.all(-0.5 <= samples[:, 0])
    assert np.all(samples[:, 0] < 0.5)
    assert np.all(-1 <= samples[:, 1])
    assert np.all(samples[:, 1] < 1)

    samples = make_uniform(n_samples=10, region=Polygon(((0, 10), (1, 11), (1, 10))), seed=rng)
    assert len(samples) == 10
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)


def test_simulate_uniform():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_uniform(n_samples=10, region=(0, 1), seed=rng)
    assert len(locdata) == 10
    assert locdata.dimension == 1
    assert repr(locdata.region) == 'Interval(0, 1)'


def test_make_Poisson():
    rng = np.random.default_rng(seed=1)

    samples = make_Poisson(intensity=10, region=EmptyRegion(), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples = make_Poisson(intensity=1e-10, region=(0, 1), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples = make_Poisson(intensity=1e-10, region=((0, 1), (10, 11)), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 3

    samples = make_Poisson(intensity=10, region=(0, 1), seed=rng)
    assert samples.shape[1] == 1
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    samples = make_Poisson(intensity=10, region=((0, 1), (10, 11)), seed=rng)
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples = make_Poisson(intensity=10, region=((0, 1), (10, 11), (100, 101)), seed=rng)
    assert samples.shape[1] == 3
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples = make_Poisson(intensity=10, region=Interval(0, 1), seed=rng)
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    samples = make_Poisson(intensity=10, region=Rectangle((0, 10), 1, 1, 0), seed=rng)
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples = make_Poisson(intensity=10, region=Ellipse((0, 0), 1, 1, 0), seed=rng)
    assert samples.shape[1] == 2
    assert np.all(-0.5 <= samples[:, 0])
    assert np.all(samples[:, 0] < 0.5)
    assert np.all(-0.5 <= samples[:, 1])
    assert np.all(samples[:, 1] < 0.5)

    samples = make_Poisson(intensity=10, region=Ellipse((0, 0), 1, 2, 0), seed=rng)
    assert samples.shape[1] == 2
    assert np.all(-0.5 <= samples[:, 0])
    assert np.all(samples[:, 0] < 0.5)
    assert np.all(-1 <= samples[:, 1])
    assert np.all(samples[:, 1] < 1)

    samples = make_Poisson(intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))), seed=rng)
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)


@pytest.mark.visual
def test_make_Poisson_visual():
    rng = np.random.default_rng(seed=1)

    samples = make_Poisson(intensity=1000, region=Polygon(((0, 10), (1, 11), (1, 10))), seed=rng)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    ax.axis('equal')
    plt.show()


def test_simulate_Poisson():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_Poisson(intensity=10, region=(0, 1), seed=rng)
    assert locdata.dimension == 1
    assert repr(locdata.region) == 'Interval(0, 1)'


def test_make_cluster():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_cluster(centers=10, region=EmptyRegion(), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples, labels, parent_samples, region = make_cluster(centers=0, region=(0, 1), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_cluster(centers=10, region=(0, 1), expansion_distance=1,
                                                   offspring=None,
                                                   clip=False, shuffle=False, seed=rng)
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert len(parent_samples) == 10
    assert np.all(-1 <= samples[:, 0])
    assert np.all(samples[:, 0] < 2)

    samples, labels, parent_samples, region = make_cluster(centers=(-1, 0, 0.5, 1, 2), region=(0, 1), expansion_distance=1,
                                                   offspring=None,
                                                   clip=True, shuffle=True, seed=rng)
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert len(parent_samples) == 5
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    offspring = [np.linspace(-0.1, 0.1, 11)] * 20
    samples, labels, parent_samples, region = make_cluster(centers=10, region=(0, 1), expansion_distance=1,
                                                   offspring=offspring,
                                                   clip=True, shuffle=True, seed=rng)
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert len(parent_samples) == 10
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    with pytest.raises(TypeError):
        offspring = [np.linspace(-0.1, 0.1, 11)] * 2
        make_cluster(centers=10, region=(0, 1), expansion_distance=1,
                     offspring=offspring,
                     clip=True, shuffle=True, seed=rng)

    samples, labels, parent_samples, region = make_cluster(centers=10, region=((0, 1), (10, 11), (100, 101)),
                                                   expansion_distance=0.1,
                                                   offspring=None,
                                                   clip=True, shuffle=True, seed=rng)
    assert len(samples) == len(labels)
    assert len(parent_samples) == 10
    assert samples.ndim == 2
    assert samples.shape[1] == 3
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples, labels, parent_samples, region = make_cluster(centers=10, region=Interval(0, 1), expansion_distance=1,
                                                   offspring=None,
                                                   clip=True, shuffle=True, seed=rng)
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert len(parent_samples) == 10
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    samples, labels, parent_samples, region = make_cluster(centers=10, region=Rectangle((0, 10), 1, 1, 0),
                                                           expansion_distance=1,
                                                           offspring=None,
                                                           clip=True, shuffle=True, seed=rng)
    assert len(samples) == len(labels)
    assert len(parent_samples) == 10
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_cluster(centers=((-0.5, 10), (0.5, 11), (1, 0)),
                                                           region=Rectangle((0, 10), 1, 1, 0), expansion_distance=1,
                                                           offspring=None,
                                                           clip=True, shuffle=True, seed=rng)
    assert len(samples) == len(labels)
    assert len(parent_samples) == 3
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] <= 11)

    samples, labels, parent_samples, region = make_cluster(centers=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                   expansion_distance=1,
                                                   offspring=None,
                                                   clip=True, shuffle=True, seed=rng)
    assert len(samples) == len(labels)
    assert len(parent_samples) == 10
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)


    def offspring(parent):
        angles = np.linspace(0, 360, 36)
        for angle in angles:
            circle = Ellipse(parent, 1, 1, angle)
        return circle.points

    samples, labels, parent_samples, region = make_cluster(centers=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                   expansion_distance=1,
                                                   offspring=offspring,
                                                   clip=True, shuffle=True, seed=rng)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    def offspring(parent):
        from math import sin
        return [sin(parent)]
    samples, labels, parent_samples, region = make_cluster(centers=10, region=(0, 1),
                                                   expansion_distance=1,
                                                   offspring=offspring,
                                                   clip=True, shuffle=True, seed=rng)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    offspring = [make_uniform(n_samples=20, region=Ellipse((0, 0), 0.1, 0.1), seed=rng)] * 10
    samples, labels, parent_samples, region = make_cluster(centers=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                   expansion_distance=0.1,
                                                   offspring=offspring,
                                                   clip=True, shuffle=True, seed=rng)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    offspring = [np.array([[1, 2], [1, 3], [2, 3]]),
                 np.array([]),
                 np.array([[1, 2], [1, 3], [2, 3]])
                 ]
    samples, labels, parent_samples, region = make_cluster(centers=3, region=((0, 10), (0, 10)),
                                                   expansion_distance=1,
                                                   offspring=offspring,
                                                   clip=True, shuffle=True, seed=rng)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)
    assert np.all(0 <= samples[:, 1])
    assert np.all(samples[:, 1] < 10)

    offspring = [np.array([])] * 20
    samples, labels, parent_samples, region = make_cluster(centers=3, region=((0, 10), (0, 10)),
                                                   expansion_distance=1,
                                                   offspring=offspring,
                                                   clip=True, shuffle=True, seed=rng)
    assert len(samples) == 0
    assert samples.ndim == 2


@pytest.mark.visual
def test_make_cluster_visual():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_cluster(centers=10, region=(0, 1), expansion_distance=1,
                                                   offspring=[np.linspace(-0.1, 0.1, 11)]*10,
                                                   clip=False, shuffle=True, seed=rng)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples, [0] * len(samples), alpha=0.1)
    plt.scatter(parent_samples, [0] * len(parent_samples), color='Red', alpha=0.1)
    ax.axis('equal')
    plt.show()

    def offspring_points(parent):
        angles = np.linspace(0, 360, 36)
        for angle in angles:
            circle = Ellipse(parent, 1, 1, angle)
        return circle.points

    samples, labels, parent_samples, region = make_cluster(centers=5, region=Polygon(((0, 10), (10, 20), (10, 10))),
                                                       offspring=offspring_points,
                                                       clip=True, shuffle=True, seed=rng)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    ax.axis('equal')
    plt.show()


def test_simulate_cluster():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_cluster(centers=10, region=(0, 1),
                               expansion_distance=1, offspring=None,
                               clip=False, shuffle=False, seed=rng)
    assert locdata.dimension == 1
    assert repr(locdata.region) == 'Interval(-1, 2)'
    assert 'cluster_label' in locdata.data.columns


def test_make_NeymanScott():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=10, region=EmptyRegion(), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=1e-10, region=(0, 1), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=10, region=(0, 1),
                                                       expansion_distance=1,
                                                       offspring=None,
                                                       clip=False, shuffle=False, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-1 <= samples[:, 0])
    assert np.all(samples[:, 0] < 2)

    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=10, region=(0, 1),
                                                       expansion_distance=0,
                                                       offspring=None,
                                                       clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    offspring = [np.linspace(-0.1, 0.1, 11)] * 200
    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=10, region=(0, 1),
                                                       expansion_distance=1,
                                                       offspring=offspring,
                                                       clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-1 <= samples[:, 0])
    assert np.all(samples[:, 0] < 2)

    with pytest.raises(TypeError):
        offspring = [np.linspace(-0.1, 0.1, 11)] * 2
        make_NeymanScott(parent_intensity=10, region=(0, 1),
                         expansion_distance=1,
                         offspring=offspring,
                         clip=True, shuffle=True, seed=rng)

    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=10,
                                                               region=((0, 1), (10, 11), (100, 101)),
                                                               expansion_distance=1,
                                                               offspring=None,
                                                               clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 3
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=10, region=Interval(0, 1),
                                                       expansion_distance=1,
                                                       offspring=None,
                                                       clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=10, region=Rectangle((0, 10), 1, 1, 0),
                                                       expansion_distance=1,
                                                       offspring=None,
                                                       clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                       expansion_distance=1,
                                                       offspring=None,
                                                       clip=False, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.shape[1] == 2
    assert samples.ndim == 2
    assert np.all(-1 <= samples[:, 0])
    assert np.all(samples[:, 0] < 2)
    assert np.all(9 <= samples[:, 1])
    assert np.all(samples[:, 1] < 12)

    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                       expansion_distance=1,
                                                       offspring=None,
                                                       clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.shape[1] == 2
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    def offspring(parent):
        angles = np.linspace(0, 360, 36)
        for angle in angles:
            circle = Ellipse(parent, 1, 1, angle)
        return circle.points
    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=10,
                                                       region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                       expansion_distance=1,
                                                       offspring=offspring,
                                                       clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.shape[1] == 2
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    def offspring(parent):
        from math import sin
        return [sin(parent)]
    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=10, region=(0, 1),
                                                       expansion_distance=1,
                                                       offspring=offspring,
                                                       clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.shape[1] == 1
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)

    offspring = [np.array([[1, 2], [1, 3], [2, 3]]),
                 np.array([]),
                 np.array([[1, 2], [1, 3], [2, 3]])
                 ] * 20
    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=0.1, region=((0, 10), (0, 10)),
                                                       expansion_distance=1,
                                                       offspring=offspring,
                                                       clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.shape[1] == 2
    assert samples.ndim == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    offspring = [np.array([])] * 20
    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=0.1, region=((0, 10), (0, 10)),
                                                       expansion_distance=1,
                                                       offspring=offspring,
                                                       clip=True, shuffle=True, seed=rng)
    assert len(samples) == 0
    assert samples.ndim == 2


@pytest.mark.visual
def test_make_NeymanScott_visual():
    rng = np.random.default_rng(seed=1)

    p_intensity = 0.5
    region = Polygon(((0, 10), (10, 20), (10, 10)))
    def offspring_points(parent):
        angles = np.linspace(0, 360, 36)
        for angle in angles:
            circle = Ellipse(parent, 1, 1, angle)
        return circle.points

    samples, labels, parent_samples, region = make_NeymanScott(parent_intensity=p_intensity, region=region,
                                                       expansion_distance=1,
                                                       offspring=offspring_points,
                                                       clip=True, shuffle=True, seed=rng)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    ax.axis('equal')
    plt.show()


def test_simulate_NeymanScott():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_NeymanScott(parent_intensity=10, region=(0, 1),
                                                       expansion_distance=0,
                                                       offspring=None,
                                                       clip=False, shuffle=True, seed=rng)
    assert locdata.dimension == 1
    assert repr(locdata.region) == 'Interval(0, 1)'
    assert 'cluster_label' in locdata.data.columns


def test_make_Matern():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_Matern(parent_intensity=10, region=EmptyRegion(), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples, labels, parent_samples, region = make_Matern(parent_intensity=1e-10, region=(0, 10), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_Matern(parent_intensity=1e-10, region=((0, 1), (10, 11)), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_Matern(parent_intensity=10, region=(0, 10),
                                                  cluster_mu=10, radius=1.0,
                                                  clip=False, shuffle=False, seed=rng)
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 12)

    samples, labels, parent_samples, region = make_Matern(parent_intensity=10, region=(0, 10),
                                                  cluster_mu=1, radius=1.0,
                                                  clip=False, shuffle=False, seed=rng)
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 12)

    samples, labels, parent_samples, region = make_Matern(parent_intensity=10, region=(0, 10),
                                                  cluster_mu=10, radius=np.linspace(1, 2, 200),
                                                  clip=False, shuffle=False, seed=rng)
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-4 <= samples[:, 0])
    assert np.all(samples[:, 0] < 14)

    samples, labels, parent_samples, region = make_Matern(parent_intensity=10, region=(0, 10),
                                                  cluster_mu=10, radius=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    samples, labels, parent_samples, region = make_Matern(parent_intensity=10, region=((0, 1), (10, 11)),
                                                  cluster_mu=10, radius=1,
                                                  clip=False, shuffle=True, seed=rng)
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    with pytest.raises(NotImplementedError):
        samples, labels, parent_samples, region = make_Matern(parent_intensity=10, region=((0, 1), (10, 11), (100, 101)),
                                                      cluster_mu=10, radius=1,
                                                      clip=False, shuffle=True, seed=rng)
        assert len(samples) == len(labels)
        assert samples.ndim == 2
        assert samples.shape[1] == 3
        assert np.all(-1 <= samples[:, 0])
        assert np.all(samples[:, 0] < 2)
        assert np.all(9 <= samples[:, 1])
        assert np.all(samples[:, 1] < 12)
        assert np.all(99 <= samples[:, 1])
        assert np.all(samples[:, 1] < 102)


    samples, labels, parent_samples, region = make_Matern(parent_intensity=10, region=Interval(0, 1),
                                                  cluster_mu=10, radius=1,
                                                  clip=False, shuffle=True, seed=rng)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)

    samples, labels, parent_samples, region = make_Matern(parent_intensity=10, region=Rectangle((0, 10), 1, 1, 0),
                                                  cluster_mu=10, radius=1,
                                                  clip=False, shuffle=True, seed=rng)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_Matern(parent_intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  cluster_mu=10, radius=1,
                                                  clip=False, shuffle=True, seed=rng)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_Matern(parent_intensity=100, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  cluster_mu=10, radius=1,
                                                  clip=True, shuffle=True, seed=rng)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)


@pytest.mark.visual
def test_make_Matern_visual():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_Matern(parent_intensity=1, region=(0, 10),
                                                  cluster_mu=10, radius=1.0,
                                                  clip=False, shuffle=False, seed=rng)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples, [0] * len(samples), alpha=0.1)
    ax.axis('equal')
    plt.show()

    samples, labels, parent_samples, region = make_Matern(parent_intensity=0.1,
                                                  region=Polygon(((0, 10), (10, 20), (10, 10))),
                                                  cluster_mu=200, radius=np.linspace(0.1, 3, 100),
                                                  clip=False, shuffle=False, seed=rng)
    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    ax.axis('equal')
    plt.show()

    samples, labels, parent_samples, region = make_Matern(parent_intensity=0.1,
                                                  region=Polygon(((0, 10), (10, 20), (10, 10))),
                                                  cluster_mu=200, radius=1,
                                                  clip=True, shuffle=True, seed=rng)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    ax.axis('equal')
    plt.show()


def test_simulate_Matern():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_Matern(parent_intensity=10, region=(0, 10),
                                                  cluster_mu=10, radius=1.0,
                                                  clip=False, shuffle=False, seed=rng)
    assert locdata.dimension == 1
    assert repr(locdata.region) == 'Interval(-1.0, 11.0)'
    assert 'cluster_label' in locdata.data.columns


def test_make_Thomas():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=EmptyRegion(), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=0, region=(0, 10), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=(0, 10), cluster_mu=0, seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=1e-10, region=(0, 10), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=(0, 10),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=False, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 16)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=(0, 10),
                                                  expansion_factor=0,
                                                  cluster_mu=0.1, cluster_std=1.0,
                                                  clip=True, shuffle=False, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 16)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=np.float64(10), region=(0, 10),
                                                  expansion_factor=np.float64(1),
                                                  cluster_mu=np.float64(10), cluster_std=np.float64(1),
                                                  clip=True, shuffle=False, seed=rng)
    assert len(samples) > 0

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=(0, 10),
                                                  expansion_factor=1,
                                                  cluster_mu=np.linspace(1, 10, 1000), cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 16)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    with pytest.raises(ValueError):
        samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=(0, 10),
                                                      expansion_factor=1,
                                                      cluster_mu=np.linspace(1, 10, 10), cluster_std=1.0,
                                                      clip=True, shuffle=True, seed=rng)

    with pytest.raises(TypeError):
        samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=(0, 10),
                                                      expansion_factor=6,
                                                      cluster_mu=10, cluster_std=(1, 2),
                                                      clip=True, shuffle=True, seed=rng)

    with pytest.raises(TypeError):
        samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=(0, 10),
                                                      expansion_factor=6,
                                                      cluster_mu=10, cluster_std=np.linspace(0.2, 2, 200),
                                                      clip=True, shuffle=True, seed=rng)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=((0, 1), (10, 11)),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=((0, 1), (10, 11), (100, 101)),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 3
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(94 <= parent_samples[:, 2])
    assert np.all(parent_samples[:, 2] < 107)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=Interval(0, 1),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=Rectangle((0, 10), 1, 1, 0),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  expansion_factor=0,
                                                  cluster_mu=np.linspace(10, 20, 200), cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  expansion_factor=0,
                                                  cluster_mu=10, cluster_std=(0.1, 1),
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 1)
    assert np.all(10 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 11)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  expansion_factor=0,
                                                  cluster_mu=10, cluster_std=np.linspace((0.01, 0.02), (0.1, 0.2), 100),
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 1)
    assert np.all(10 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 11)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)


@pytest.mark.visual
def test_make_Thomas_visual():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=0.2, region=(0, 10),
                                                  expansion_factor=6,
                                                  cluster_mu=100, cluster_std=1.0,
                                                  clip=True, shuffle=False, seed=rng)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples, [0] * len(samples), alpha=0.1)
    plt.scatter(parent_samples, [0] * len(parent_samples), c='Red', alpha=0.2)
    ax.axis('equal')
    plt.show()

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=50, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  expansion_factor=0,
                                                  cluster_mu=100, cluster_std=np.linspace((0.001, 0.002), (0.02, 0.04), 100),
                                                  clip=True, shuffle=False, seed=rng)
    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    plt.scatter(parent_samples[:, 0], parent_samples[:, 1], c='Red')
    ax.axis('equal')
    plt.show()

    samples, labels, parent_samples, region = make_Thomas(parent_intensity=20,
                                                  region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  expansion_factor=2,
                                                  cluster_mu=500, cluster_std=0.02,
                                                  clip=True, shuffle=True, seed=rng)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    plt.scatter(samples[:, 0], samples[:, 1])
    plt.scatter(parent_samples[:, 0], parent_samples[:, 1], c='Red')
    ax.axis('equal')
    plt.show()


def test_simulate_Thomas():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_Thomas(parent_intensity=10, region=(0, 10),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=False, seed=rng)
    assert locdata.dimension == 1
    assert repr(locdata.region) == 'Interval(0, 10)'
    assert 'cluster_label' in locdata.data.columns


def test_make_dstorm():
    rng = np.random.default_rng(seed=1)

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=EmptyRegion(), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 1

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=0, region=(0, 10), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=1e-10, region=(0, 10), seed=rng)
    assert np.size(samples) == 0
    assert samples.ndim == 2

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=(0, 10),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=False, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 16)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=np.float64(10), region=(0, 10),
                                                  expansion_factor=np.float64(1),
                                                  cluster_mu=np.float64(10), cluster_std=np.float64(1),
                                                  clip=True, shuffle=False, seed=rng)
    assert len(samples) > 0

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=(0, 10),
                                                  expansion_factor=1,
                                                  cluster_mu=np.linspace(1, 10, 1000), cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 16)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 10)

    with pytest.raises(ValueError):
        samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=(0, 10),
                                                      expansion_factor=1,
                                                      cluster_mu=np.linspace(1, 10, 10), cluster_std=1.0,
                                                      clip=True, shuffle=True, seed=rng)

    with pytest.raises(TypeError):
        samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=(0, 10),
                                                      expansion_factor=6,
                                                      cluster_mu=10, cluster_std=(1, 2),
                                                      clip=True, shuffle=True, seed=rng)

    with pytest.raises(TypeError):
        samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=(0, 10),
                                                      expansion_factor=6,
                                                      cluster_mu=10, cluster_std=np.linspace(0.2, 2, 200),
                                                      clip=True, shuffle=True, seed=rng)

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=((0, 1), (10, 11)),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=((0, 1), (10, 11), (100, 101)),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 3
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(94 <= parent_samples[:, 2])
    assert np.all(parent_samples[:, 2] < 107)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)
    assert np.all(100 <= samples[:, 2])
    assert np.all(samples[:, 2] < 101)

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=Interval(0, 1),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=Rectangle((0, 10), 1, 1, 0),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(-2 <= samples[:, 0])
    assert np.all(samples[:, 0] < 3)
    assert np.all(8 <= samples[:, 1])
    assert np.all(samples[:, 1] < 13)

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  expansion_factor=0,
                                                  cluster_mu=np.linspace(10, 20, 200), cluster_std=1.0,
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(-6 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 7)
    assert np.all(4 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 17)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  expansion_factor=0,
                                                  cluster_mu=10, cluster_std=(0.1, 1),
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 1)
    assert np.all(10 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 11)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=Polygon(((0, 10), (1, 11), (1, 10))),
                                                  expansion_factor=0,
                                                  cluster_mu=10, cluster_std=np.linspace((0.01, 0.02), (0.1, 0.2), 100),
                                                  clip=True, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert samples.ndim == 2
    assert samples.shape[1] == 2
    assert np.all(0 <= parent_samples[:, 0])
    assert np.all(parent_samples[:, 0] < 1)
    assert np.all(10 <= parent_samples[:, 1])
    assert np.all(parent_samples[:, 1] < 11)
    assert np.all(0 <= samples[:, 0])
    assert np.all(samples[:, 0] < 1)
    assert np.all(10 <= samples[:, 1])
    assert np.all(samples[:, 1] < 11)

    samples, labels, parent_samples, region = make_dstorm(parent_intensity=10, region=(0, 10),
                                                  expansion_factor=0,
                                                  cluster_mu=10, min_points=5, cluster_std=0.1,
                                                  clip=False, shuffle=True, seed=rng)
    assert len(samples) > 0
    assert len(samples) == len(labels)
    assert samples.ndim == 2
    assert samples.shape[1] == 1
    assert np.bincount(labels).mean() == pytest.approx(10, rel=0.2)


def test_simulate_dstorm():
    rng = np.random.default_rng(seed=1)
    locdata = simulate_dstorm(parent_intensity=10, region=(0, 10),
                                                  expansion_factor=6,
                                                  cluster_mu=10, cluster_std=1.0,
                                                  clip=True, shuffle=False, seed=rng)
    assert locdata.dimension == 1
    assert repr(locdata.region) == 'Interval(0, 10)'
    assert 'cluster_label' in locdata.data.columns

@pytest.fixture()
def locdata_simple():
    localization_dict = {
        'position_x': [0, 0, 1, 4, 5],
        'position_y': [0, 1, 3, 4, 1],
        'position_z': [0, 1, 3, 4, 1],
        'intensity': [0, 1, 3, 4, 1],
        'uncertainty_x': [10, 30, 100, 300, 0],
        'uncertainty_y': [10, 30, 100, 300, 10],
        'uncertainty_z': [10, 30, 100, 300, 10],
        }
    return LocData(dataframe=pd.DataFrame.from_dict(localization_dict))


def test_resample(locdata_simple):
    dat = resample(locdata=locdata_simple, n_samples=3)
    # print(dat.data)
    assert len(dat) == 15


def test_simulate_tracks():
    dat = simulate_tracks(n_walks=2, n_steps=3)
    # print(dat.data)
    # print(dat.meta)
    assert len(dat) == 6
    assert len(dat.coordinate_labels) == 2


def test__random_walk_drift():
    cumsteps = _random_walk_drift(n_steps=10, diffusion_constant=(1, 10), velocity=(0, 0), seed=1)
    assert cumsteps.shape == (2, 10)
    cumsteps = _random_walk_drift(n_steps=10, diffusion_constant=(0, 0), velocity=(1, 10), seed=1)
    assert cumsteps.shape == (2, 10)


def test__drift():
    frames = np.arange(2, 10, 2)
    n_frames = len(frames)
    position_deltas = _drift(frames, diffusion_constant=(1, 10), velocity=None, seed=1)
    assert position_deltas.shape == (2, n_frames)
    position_deltas = _drift(frames, diffusion_constant=None, velocity=(1, 2), seed=1)
    assert position_deltas.shape == (2, n_frames)
    position_deltas = _drift(frames, diffusion_constant=None, velocity=None, seed=1)
    assert position_deltas is None


def test_simulate_drift(locdata_2d):
    new_locdata = add_drift(locdata_2d, diffusion_constant=None, velocity=None, seed=1)
    assert len(new_locdata) == len(locdata_2d)
    new_locdata = add_drift(locdata_2d, diffusion_constant=(1, 10), velocity=(10, 10), seed=1)
    assert len(new_locdata) == len(locdata_2d)
    # print(new_locdata.meta)


@pytest.mark.visual
def test_visual__drift():
    frames = np.arange(0, 1_000_000, dtype=int)
    print(frames.shape)

    # position_deltas = _drift(frames=frames, velocity=(1, 2), seed=1)
    position_deltas = _drift(frames=frames, diffusion_constant=(1, 2), seed=1)
    print(position_deltas.shape)
    print(position_deltas[0, :10])
    for pd in position_deltas:
        plt.plot(frames, pd)
        plt.plot(frames, pd)
    plt.show()


@pytest.mark.visual
def test_visual_add_drift(locdata_2d):
    # new_locdata = add_drift(locdata_2d, diffusion_constant=(1, 10), velocity=(10, 10), seed=1)
    new_locdata = add_drift(locdata_2d, diffusion_constant=None, velocity=(1, 1), seed=1)
    ax = locdata_2d.data.plot(*locdata_2d.coordinate_labels, kind='scatter')
    new_locdata.data.plot(*new_locdata.coordinate_labels, kind='scatter', ax=ax, c='r')
    plt.show()


def test_simulate_frame_numbers(locdata_2d):
    frames = simulate_frame_numbers(n_samples=(len(locdata_2d)), lam=2)
    assert len(frames) == len(locdata_2d)



from io import StringIO

import locan.constants
from locan.locan_io import load_Nanoimager_file
from locan.locan_io.locdata.nanoimager_io import load_Nanoimager_header


def test_get_correct_column_names_from_Nanoimager_header():
    columns = load_Nanoimager_header(
        path=locan.ROOT_DIR / 'tests/test_data/Nanoimager_dstorm_data.csv')
    assert (columns == ['channel', 'frame', 'position_x', 'position_y', 'position_z', 'intensity',
                        'local_background'])

    file_like = StringIO("Channel,Frame,X (nm),Y (nm),Z (nm),Photons,Background\n"
                         "0,1548,40918.949219,56104.691406,0.000000,139.828232,0.848500")
    columns = load_Nanoimager_header(path=file_like)
    assert (columns == ['channel', 'frame', 'position_x', 'position_y', 'position_z', 'intensity',
                        'local_background'])


def test_loading_Nanoimager_file():
    dat = load_Nanoimager_file(path=locan.ROOT_DIR / 'tests/test_data/Nanoimager_dstorm_data.csv',
                                  nrows=10)
    #print(dat.data.columns)
    assert (len(dat) == 10)
    assert all(dat.data.columns == ['channel', 'frame', 'position_x', 'position_y', 'position_z', 'intensity',
                                    'local_background'])

    file_like = StringIO("Channel,Frame,X (nm),Y (nm),Z (nm),Photons,Background\n"
                         "0,1548,40918.949219,56104.691406,0.000000,139.828232,0.848500")
    dat = load_Nanoimager_file(path=file_like, nrows=1)
    assert (len(dat) == 1)
    assert all(dat.data.columns == ['channel', 'frame', 'position_x', 'position_y', 'position_z', 'intensity',
                                    'local_background'])



"""
Transform intensity values.

This module provides functions and classes to rescale intensity values.
In addition to the presented functions, rescaling can further be performed by third-party modules like:
1) matplotlib.colors
2) skimage.exposure
3) astropy.visualization.

Callable transformation classes that inherit from :class:`matplotlib.colors.Normalize` and specify an inverse transformation
can be passed to the `norm` parameter.

"""

from abc import ABC, abstractmethod
from enum import Enum

import numpy as np
import matplotlib.colors as mcolors
from skimage import exposure


__all__ = ['Trafo', 'HistogramEqualization', 'adjust_contrast']


class Trafo(Enum):
    """
    Standard definitions for intensity transformation.
    """
    NONE = 0
    """no transformation"""
    STANDARDIZE = 1
    """rescale (min, max) to (0, 1)"""
    STANDARDIZE_UINT8 = 2
    """rescale (min, max) to (0, 255)"""
    ZERO = 3
    """rescale (0, max) to (0, 1)"""
    ZERO_UINT8 = 4
    """rescale (0, max) to (0, 255)"""
    EQUALIZE = 5
    """equalize histogram from all values > 0"""
    EQUALIZE_UINT8 = 6
    """equalize histogram from all values > 0 onto (0, 255)"""
    EQUALIZE_ALL = 7
    """equalize histogram"""
    EQUALIZE_ALL_UINT8 = 8
    """equalize histogram onto (0, 255)"""
    EQUALIZE_0P3 = 9
    """equalize histogram with power = 0.3 from all values > 0"""
    EQUALIZE_0P3_UINT8 = 10
    """equalize histogram with power = 0.3 from all values > 0 onto (0, 255)"""
    EQUALIZE_0P3_ALL = 11
    """equalize histogram with power = 0.3"""
    EQUALIZE_0P3_ALL_UINT8 = 12
    """equalize histogram with power = 0.3 onto (0, 255)"""


class Transform(ABC):
    """
    Abstract base class for transformation classes.
    """

    @abstractmethod
    def __call__(self, values, clip=True):
        """
        Transform values.

        Parameters
        ----------
        values : array-like
            The input values
        clip : bool
            If `True` values outside the [0:1] range are
            clipped to the [0:1] range.

        Returns
        -------
        numpy.ndarray
            The transformed values.
        """
        raise NotImplementedError

        _values, is_scalar = self.process_value(values)
        self.autoscale_None(_values)  # sets self.vmin, self.vmax if None

        if clip is None:
            clip = self.clip

        # Normalize based on vmin and vmax
        np.subtract(_values, self.vmin, out=_values)
        np.true_divide(_values, self.vmax - self.vmin, out=_values)

        # Clip to the 0 to 1 range
        if clip:
            np.clip(_values, 0., 1., out=_values)

        new_values =_values
        return new_values

    @property
    @abstractmethod
    def inverse(self):
        """A transformation that performs the inverse operation."""
        raise NotImplementedError


class HistogramEqualization(mcolors.Normalize, Transform):
    """
    Histogram equalization with power intensification.

    The transformation function as described in [1]_ is :math:`f(a, p)` according to:

    .. math::

       \\frac{f(a) - f(a_min)} {f(a_max) - f(a_min)} =
       \\frac{\\int_{a_min}^{a}{h^p(a') da'}} {\\int_{a_min}^{a_max}{h^p(a') da'}}

    Here, :math:`a` is an intensity value, :math:`p` the power (a parameter), and :math:`h(a)` the histogram of
    intensities.

    Notes
    -----
    The default for n_bins is 65536 (16 bit).
    For most SMLM datasets this should be sufficient to resolve individual localizations despite a large dynamic range.
    Setting n_bins to 256 (8 bit) is too course for many SMLM datasets.

    References
    ----------
    .. [1] Yaroslavsky, L. (1985) Digital picture processing. Springer, Berlin.

    Parameters
    ----------
    reference : array-like
        The data values to define the transformation function. If None then the values in `__call__` are used.
    power : float
        The `power` intensification parameter.
    n_bins : int
        Number of bins used to compute the intensity histogram.
    mask : array-like[bool]
        A bool mask with shape equal to that of values. If reference is None, reference is set to values[mask].
        The transformation determined from reference is then applied to all values.
    """

    def __init__(self, vmin=None, vmax=None, reference=None, power=1, n_bins=65536, mask=None):
        super().__init__(vmin=vmin, vmax=vmax)
        self.reference = reference
        self.power = power
        self.n_bins = n_bins
        self.mask = mask

    def __call__(self, values):
        """
        Histogram equalization with power intensification.

        Parameters
        ----------
        values : array-like
            The input values.

        Returns
        -------
        numpy.ndarray
        """
        _values, is_scalar = self.process_value(values)
        self.autoscale_None(_values)  # sets self.vmin, self.vmax if None

        # Normalize based on vmin and vmax
        np.subtract(_values, self.vmin, out=_values)
        np.true_divide(_values, self.vmax - self.vmin, out=_values)

        # Clip to the 0 to 1 range
        np.clip(_values, 0., 1., out=_values)

        if self.reference is None:
            _reference = _values[self.mask]
        else:
            _reference = self.reference

        data, bin_edges = np.histogram(_reference, bins=self.n_bins, range=(0, 1))
        bin_centers = bin_edges[:-1] + np.diff(bin_edges) / 2
        cdf = np.cumsum(data**self.power)
        cdf = cdf / cdf[-1]
        new_values = np.interp(_values, bin_centers, cdf)
        return new_values

    @property
    def inverse(self, values):
        """A Transformation object that performs the inverse operation."""
        raise NotImplementedError


def adjust_contrast(image, rescale=True, **kwargs):
    """
    Adjust contrast of image by a predefined transformation:

    Parameters
    ----------
    image : array-like
        Values to be adjusted
    rescale : int, str, Trafo, callable, bool, None
        Transformation as defined in :class:`locan.constants.Trafo` or by transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min intensities
        For 'equal' intensity values are rescaled by histogram equalization.
    kwargs : dict
        Other parameters that are passed to the specific Transformation class.

    Returns
    -------
    numpy.ndarray
    """
    if rescale is None or rescale is False or rescale is Trafo.NONE or rescale == 0:
        new_image = image

    elif rescale is True or rescale == 1 or rescale is Trafo.STANDARDIZE \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.STANDARDIZE.name):
        new_image = exposure.rescale_intensity(image * 1., **kwargs)

    elif rescale == 2 or rescale is Trafo.STANDARDIZE_UINT8 \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.STANDARDIZE_UINT8.name):
        new_image = exposure.rescale_intensity(image,
                                               **dict(dict(out_range=(0, 255)), **kwargs)
                                               ).astype(np.uint8)

    elif rescale == 3 or rescale is Trafo.ZERO \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.ZERO.name):
        new_image = exposure.rescale_intensity(image * 1.,
                                               **dict(dict(in_range=(0, np.nanmax(image)),
                                                           out_range=(0, 1)), **kwargs)
                                               )

    elif rescale == 4 or rescale is Trafo.ZERO_UINT8 \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.ZERO_UINT8.name):
        new_image = exposure.rescale_intensity(image,
                                               **dict(dict(in_range=(0, np.nanmax(image)),
                                                           out_range=(0, 255)), **kwargs)
                                               ).astype(np.uint8)

    elif rescale == 5 or rescale is Trafo.EQUALIZE \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.EQUALIZE.name):
        norm = HistogramEqualization(**dict(dict(power=1, mask=image > 0), **kwargs))
        new_image = norm(image)

    elif rescale == 6 or rescale is Trafo.EQUALIZE_UINT8 \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.EQUALIZE_UINT8.name):
        norm = HistogramEqualization(**dict(dict(power=1, mask=image > 0), **kwargs))
        new_image = np.multiply(norm(image), 255).astype(np.uint8)

    elif rescale == 7 or rescale is Trafo.EQUALIZE_ALL \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.EQUALIZE_ALL.name):
        norm = HistogramEqualization(**dict(dict(power=1, mask=None), **kwargs))
        new_image = norm(image)

    elif rescale == 8 or rescale is Trafo.EQUALIZE_ALL_UINT8 \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.EQUALIZE_ALL_UINT8.name):
        norm = HistogramEqualization(**dict(dict(power=1, mask=None), **kwargs))
        new_image = np.multiply(norm(image), 255).astype(np.uint8)

    elif rescale == 9 or rescale is Trafo.EQUALIZE_0P3 \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.EQUALIZE_0P3.name):
        norm = HistogramEqualization(**dict(dict(power=0.3, mask=image > 0), **kwargs))
        new_image = norm(image)

    elif rescale == 10 or rescale is Trafo.EQUALIZE_0P3_UINT8 \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.EQUALIZE_0P3_UINT8.name):
        norm = HistogramEqualization(**dict(dict(power=0.3, mask=image > 0), **kwargs))
        new_image = np.multiply(norm(image), 255).astype(np.uint8)

    elif rescale == 11 or rescale is Trafo.EQUALIZE_0P3_ALL \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.EQUALIZE_0P3_ALL.name):
        norm = HistogramEqualization(**dict(dict(power=0.3, mask=None), **kwargs))
        new_image = norm(image)

    elif rescale == 12 or rescale is Trafo.EQUALIZE_0P3_ALL_UINT8 \
            or (isinstance(rescale, str) and rescale.upper() == Trafo.EQUALIZE_0P3_ALL_UINT8.name):
        norm = HistogramEqualization(**dict(dict(power=0.3, mask=None), **kwargs))
        new_image = np.multiply(norm(image), 255).astype(np.uint8)

    # to be deprecated eventually
    elif rescale == 'equal':
        new_image = exposure.equalize_hist(image, **kwargs)
    elif rescale == 'unity':
        new_image = exposure.rescale_intensity(image * 1., **kwargs)
    elif isinstance(rescale, tuple):
        p_low, p_high = np.ptp(image) * np.asarray(rescale) / 100 + image.min()
        new_image = exposure.rescale_intensity(image, in_range=(p_low, p_high))
    elif callable(rescale):
        new_image = rescale(image, **kwargs)
    else:
        raise TypeError('Transformation is not defined.')

    return new_image
